import logging

import numpy as np
from fibsem.microscope import FibsemMicroscope
from fibsem import calibration
from fibsem.structures import (
    BeamSystemSettings,
    BeamType,
    ImageSettings,
    MicroscopeSettings,
    MicroscopeState,
    FibsemImage,
)


# TODO: change return type to list of warnings rather than reading the log...
def validate_initial_microscope_state(
    microscope: FibsemMicroscope, settings: MicroscopeSettings
) -> None:
    """Set the initial microscope state to default, and validate other settings."""

    # validate chamber state
    _validate_chamber_state(microscope=microscope)

    # validate stage calibration (homed, linked)
    _validate_stage_calibration(microscope=microscope)

    # validate needle calibration (needle calibration, retracted)
    _validate_needle_calibration(microscope=microscope)

    # validate beam settings and calibration
    _validate_beam_system_settings(microscope=microscope, settings=settings)

    # validate scan rotation
    microscope.set("scan_rotation", value=0.0, beam_type=BeamType.ION)



def _validate_stage_calibration(microscope: FibsemMicroscope) -> None:
    """Validate if the required stage calibration has been performed.

    Args:
        microscope (FibsemMicroscope): autoscript microscope instance
    """
    # QUERY: should we home?
    if not microscope.get("stage_homed"):
        logging.warning("Stage is not homed.")

    # QUERY: should we focus and link?
    if not microscope.get("stage_linked"):
        logging.warning("Stage is not linked.")

    logging.info("Stage calibration validation complete.")

    return


def _validate_needle_calibration(microscope: FibsemMicroscope) -> None:
    """Validate if the needle is inserted
    Args:
        microscope (FibsemMicroscope): autoscript microscope instance
    """

    needle_state = microscope.get("manipulator_state")
    logging.info(f"Needle is {needle_state}")

    if needle_state != "Retracted":
        logging.warning(f"Needle is {needle_state}. Please retract before starting.")

    # movement.retract_needle() # TODO: decide whether or not to do this
    # TODO: calibrate needle? check if needle has been calibrated? how

    return


# TODO: use _set_type_mode for setting the detector type and mode not directly setting the values


def _validate_beam_system_state(
    microscope: FibsemMicroscope, settings: BeamSystemSettings
) -> None:

    beam_name = settings.beam_type.name

    logging.info(f"Validating {beam_name} Beam")
    if not microscope.get("on", settings.beam_type):
        logging.warning(f"{beam_name} Beam is not on, switching on now...")
        microscope.set("on", True, settings.beam_type)
        assert microscope.get("on", settings.beam_type), f"Unable to turn on {beam_name} Beam."
        logging.warning(f"{beam_name} Beam turned on.")

    # blanked?
    if microscope.get("blanked", settings.beam_type):
        logging.warning(f"{beam_name} Beam is blanked, unblanking now...")
        microscope.set("blanked", False, settings.beam_type)
        assert not microscope.get("blanked", settings.beam_type), f"Unable to unblank {beam_name} Beam."
        logging.warning(f"{beam_name} Beam unblanked.")

    # set detectors
    microscope.set("detector_type", settings.detector_type, settings.beam_type)
    microscope.set("detector_mode", settings.detector_mode, settings.beam_type)       

    # validate working distances
    if not check_working_distance_is_within_tolerance(
        microscope, settings, atol=0.5e-3
    ):
        logging.warning(
            f"""{beam_name} Beam is not close to eucentric height. It should be {settings.eucentric_height}m
            (Currently is {microscope.get("working_distance", settings.beam_type):.4f}m)"""
        )

    # validate high voltage
    high_voltage_limits = str(microscope.get("voltage_limits", settings.beam_type))
    logging.info(f"{beam_name} Beam High Voltage Limits are: {high_voltage_limits}")

    if microscope.get("voltage", settings.beam_type) != settings.voltage:
        logging.warning(
            f"{beam_name} Beam High Voltage should be {settings.voltage}V (Currently {microscope.get('voltage', settings.beam_type)}V)"
        )

        if bool(microscope.get("voltage_controllable", settings.beam_type)):
            logging.warning(
                f"Changing {beam_name} Beam High Voltage to {settings.voltage}V."
            )
            microscope.set("voltage", settings.voltage, settings.beam_type)
            assert (
                microscope.get("voltage", settings.beam_type) == settings.voltage
            ), f"Unable to change {beam_name} Beam High Voltage"
            logging.warning(f"{beam_name} Beam High Voltage Changed")

    # validate plasma gas (only for ION beam)
    if settings.beam_type is BeamType.ION:

        plasma_gas = settings.plasma_gas.capitalize()
        if plasma_gas not in microscope.get_available(key="plasma_gas", beam_type=settings.beam_type):
            logging.warning(f"{plasma_gas} is not available as a plasma gas.")

        current_plasma_gas = microscope.get("plasma_gas", settings.beam_type)
        if current_plasma_gas != plasma_gas:
            logging.warning(f"Plasma Gas is should be {plasma_gas} (Currently {current_plasma_gas})")

def _validate_beam_system_settings(
    microscope: FibsemMicroscope, settings: MicroscopeSettings
) -> None:
    """Validate Beam Settings"""

    # electron beam
    _validate_beam_system_state(microscope, settings.system.electron)

    # ion beam
    _validate_beam_system_state(microscope, settings.system.ion)


def _validate_chamber_state(microscope: FibsemMicroscope) -> None:
    """Validate the state of the chamber"""

    chamber_state = str(microscope.get('chamber_state'))
    chamber_pressure = microscope.get("chamber_pressure")
    
    logging.info(f"Vacuum Chamber State: {chamber_state}")
    if not chamber_state == "Pumped":
        logging.warning(
            f"Chamber vacuum state should be Pumped (Currently is {chamber_state})"
        )

    logging.info(f"Vacuum Chamber Pressure: {chamber_pressure:.6f} mbar"
    )
    if chamber_pressure >= 1e-4:
        logging.warning(f"Chamber pressure is too high, please pump the system (Currently {chamber_pressure:.6f} mbar)"            )

    logging.info(f"Vacuum Chamber State Validation finished.")

def validate_stage_height_for_needle_insertion(
    microscope: FibsemMicroscope, needle_stage_height_limit: float = 3.7e-3
) -> bool:
    """Check if the needle can be inserted, based on the current stage height.

    Args:
        microscope (SdbMicroscopeClient): autoscript microscope limit
        needle_stage_height_limit (float, optional): minimum stage height limit. Defaults to 3.7e-3.

    Returns:
        bool: needle is insertable
    """

    stage_position = microscope.get_stage_position()

    # Unable to insert the needle if the stage height is below this limit (3.7e-3)
    return bool(stage_position.z > needle_stage_height_limit)


def validate_focus(
    microscope: FibsemMicroscope,
    settings: BeamSystemSettings,
    link: bool = True,
) -> bool:

    # check focus distance is within tolerance
    # if link:
    #     calibration.auto_link_stage(microscope)

    return check_working_distance_is_within_tolerance(microscope, settings=settings)


def check_working_distance_is_within_tolerance(
    microscope: FibsemMicroscope, settings: BeamSystemSettings, atol=0.5e-3
) -> bool:

    working_distance = microscope.get("working_distance", settings.beam_type)
    eucentric_height = settings.eucentric_height

    logging.info(
        f"{settings.beam_type.name} Beam working distance is {working_distance:.4f}m"
    )
    return np.isclose(working_distance, eucentric_height, atol=atol)


def check_shift_within_tolerance(
    dx: float, dy: float, ref_image: FibsemImage, limit: float = 0.25
) -> bool:
    """Check if required shift is wihtin safety limit"""
    # check if the cross correlation movement is within the safety limit
    
    pixelsize_x = ref_image.metadata.pixel_size.x
    width, height = ref_image.metadata.image_settings.resolution
    X_THRESHOLD = limit * pixelsize_x * width
    Y_THRESHOLD = limit * pixelsize_x * height

    return abs(dx) < X_THRESHOLD and abs(dy) < Y_THRESHOLD



def _validate_milling_protocol(
    stage_protocol: dict, settings: MicroscopeSettings
) -> dict:

    if "milling_current" not in stage_protocol:
        stage_protocol["milling_current"] = settings.milling.milling_current
    if "cleaning_cross_section" not in stage_protocol:
        stage_protocol["cleaning_cross_section"] = False
    if "rotation" not in stage_protocol:
        stage_protocol["rotation"] = 0.0
    if "scan_direction" not in stage_protocol:
        stage_protocol["scan_direction"] = "TopToBottom"

    # remove list element from settings
    if "protocol_stages" in stage_protocol:
        del stage_protocol["protocol_stages"]

    return stage_protocol


def _validate_configuration_values(microscope: FibsemMicroscope, dictionary: dict):
    """Recursively traverse dictionary and validate all parameters.

    Args:
        microscope (SdbMicroscopeClient): autoscript microscope instance
        dictionary (dict): settings dictionary

    Returns:
        dict: validated settings dictionary
    """
    # TODO: this function is inefficient, recursive calls can be wasteful
    for key, item in dictionary.items():
        if isinstance(item, dict):
            _validate_configuration_values(microscope, item)
        elif isinstance(item, list):
            dictionary[key] = [
                _validate_configuration_values(microscope, i)
                for i in item
                if isinstance(i, dict)
            ]
        else:
            if isinstance(item, float):
                if "hfw" in key:
                    microscope.check_available_values(key, values=[item])

                if "milling_current" in key:
                    microscope.check_available_values("current", values=[item], beam_type=BeamType.ION)
                
                if "resolution" in key:
                    microscope.check_available_values(key, values=[item])

                if "dwell_time" in key:
                    microscope.check_available_values(key, values=[item])
                    
            if isinstance(item, str):
                if "application_file" in key:
                    microscope.check_available_values("application_file", values=[item])

    return dictionary



# new validation v2

from fibsem import utils
from fibsem.structures import BeamType
from fibsem.microscope import FibsemMicroscope
import logging


def validate_microscope(microscope: FibsemMicroscope):
    # append exceptions to list
    warnings = []

    # check beams are on
    if not microscope.get("on", BeamType.ELECTRON):
        warnings.append("Electron beam is off")

    if not microscope.get("on", BeamType.ION):
        warnings.append("Ion beam is off")

    # check chamber is pumped
    if microscope.get("chamber_state") != "Pumped":
        warnings.append("Chamber is not pumped")


    # ThermoFisher specific validation
    from fibsem.microscope import ThermoMicroscope, DemoMicroscope
    if isinstance(microscope, (ThermoMicroscope, DemoMicroscope)):
        # check stage is homed
        if not microscope.get("stage_homed"):
            warnings.append("Stage is not homed")

        # check stage is linked
        if not microscope.get("stage_linked"):
            warnings.append("Stage is not linked")

        # check needle is retracted
        if microscope.get("manipulator_state") != "Retracted":
            warnings.append("Needle is not retracted")


    logging.warning(f"Microscope Validation Warnings: {warnings}")

    return warnings



from fibsem.segmentation.onnx_model import export_model_to_onnx
import argparse


def main():

    parser = argparse.ArgumentParser()
    parser.add_argument(
        "--checkpoint",
        type=str,
        help="path to openfibsem model checkpoint",
        required=True,
    )
    parser.add_argument(
        "--output",
        type=str,
        help="path to save onnx model",
        required=True,
    )
    args = parser.parse_args()
    export_model_to_onnx(args.checkpoint, args.output)




if __name__ == "__main__":
    main()


# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'ImageSettingsWidget.ui'
#
# Created by: PyQt5 UI code generator 5.15.9
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(535, 878)
        self.gridLayout = QtWidgets.QGridLayout(Form)
        self.gridLayout.setObjectName("gridLayout")
        self.pushButton_take_image = QtWidgets.QPushButton(Form)
        self.pushButton_take_image.setObjectName("pushButton_take_image")
        self.gridLayout.addWidget(self.pushButton_take_image, 29, 0, 1, 2)
        self.pushButton_take_all_images = QtWidgets.QPushButton(Form)
        self.pushButton_take_all_images.setObjectName("pushButton_take_all_images")
        self.gridLayout.addWidget(self.pushButton_take_all_images, 30, 0, 1, 2)
        self.scrollArea = QtWidgets.QScrollArea(Form)
        self.scrollArea.setMinimumSize(QtCore.QSize(0, 0))
        self.scrollArea.setWidgetResizable(True)
        self.scrollArea.setObjectName("scrollArea")
        self.scrollAreaWidgetContents = QtWidgets.QWidget()
        self.scrollAreaWidgetContents.setGeometry(QtCore.QRect(0, -46, 501, 811))
        self.scrollAreaWidgetContents.setObjectName("scrollAreaWidgetContents")
        self.gridLayout_2 = QtWidgets.QGridLayout(self.scrollAreaWidgetContents)
        self.gridLayout_2.setObjectName("gridLayout_2")
        self.label_detector_header = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        font = QtGui.QFont()
        font.setPointSize(10)
        font.setBold(True)
        font.setWeight(75)
        self.label_detector_header.setFont(font)
        self.label_detector_header.setObjectName("label_detector_header")
        self.gridLayout_2.addWidget(self.label_detector_header, 26, 0, 1, 3)
        self.label_detector_contrast = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.label_detector_contrast.setObjectName("label_detector_contrast")
        self.gridLayout_2.addWidget(self.label_detector_contrast, 30, 0, 1, 1)
        self.label_shift = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.label_shift.setObjectName("label_shift")
        self.gridLayout_2.addWidget(self.label_shift, 23, 0, 1, 1)
        self.doubleSpinBox_image_hfw = QtWidgets.QDoubleSpinBox(self.scrollAreaWidgetContents)
        self.doubleSpinBox_image_hfw.setMaximum(100000000.0)
        self.doubleSpinBox_image_hfw.setSingleStep(0.1)
        self.doubleSpinBox_image_hfw.setProperty("value", 150.0)
        self.doubleSpinBox_image_hfw.setObjectName("doubleSpinBox_image_hfw")
        self.gridLayout_2.addWidget(self.doubleSpinBox_image_hfw, 5, 1, 1, 2)
        self.label_title = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        font = QtGui.QFont()
        font.setPointSize(10)
        font.setBold(True)
        font.setWeight(75)
        self.label_title.setFont(font)
        self.label_title.setObjectName("label_title")
        self.gridLayout_2.addWidget(self.label_title, 2, 0, 1, 3)
        self.label_utils_header = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        font = QtGui.QFont()
        font.setPointSize(10)
        font.setBold(True)
        font.setWeight(75)
        self.label_utils_header.setFont(font)
        self.label_utils_header.setObjectName("label_utils_header")
        self.gridLayout_2.addWidget(self.label_utils_header, 33, 0, 1, 1)
        self.crosshair_checkbox = QtWidgets.QCheckBox(self.scrollAreaWidgetContents)
        self.crosshair_checkbox.setChecked(True)
        self.crosshair_checkbox.setObjectName("crosshair_checkbox")
        self.gridLayout_2.addWidget(self.crosshair_checkbox, 34, 1, 1, 1)
        self.checkBox_image_save_image = QtWidgets.QCheckBox(self.scrollAreaWidgetContents)
        self.checkBox_image_save_image.setObjectName("checkBox_image_save_image")
        self.gridLayout_2.addWidget(self.checkBox_image_save_image, 8, 0, 1, 3)
        self.detector_brightness_slider = QtWidgets.QSlider(self.scrollAreaWidgetContents)
        self.detector_brightness_slider.setMaximum(100)
        self.detector_brightness_slider.setSliderPosition(50)
        self.detector_brightness_slider.setOrientation(QtCore.Qt.Horizontal)
        self.detector_brightness_slider.setInvertedAppearance(False)
        self.detector_brightness_slider.setTickPosition(QtWidgets.QSlider.TicksAbove)
        self.detector_brightness_slider.setTickInterval(10)
        self.detector_brightness_slider.setObjectName("detector_brightness_slider")
        self.gridLayout_2.addWidget(self.detector_brightness_slider, 29, 1, 1, 1)
        self.label_detector_type = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.label_detector_type.setObjectName("label_detector_type")
        self.gridLayout_2.addWidget(self.label_detector_type, 27, 0, 1, 1)
        self.set_detector_button = QtWidgets.QPushButton(self.scrollAreaWidgetContents)
        self.set_detector_button.setObjectName("set_detector_button")
        self.gridLayout_2.addWidget(self.set_detector_button, 31, 0, 1, 3)
        self.detector_contrast_slider = QtWidgets.QSlider(self.scrollAreaWidgetContents)
        self.detector_contrast_slider.setMaximum(100)
        self.detector_contrast_slider.setSliderPosition(50)
        self.detector_contrast_slider.setOrientation(QtCore.Qt.Horizontal)
        self.detector_contrast_slider.setTickPosition(QtWidgets.QSlider.TicksAbove)
        self.detector_contrast_slider.setTickInterval(10)
        self.detector_contrast_slider.setObjectName("detector_contrast_slider")
        self.gridLayout_2.addWidget(self.detector_contrast_slider, 30, 1, 1, 1)
        self.label_image_hfw = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.label_image_hfw.setObjectName("label_image_hfw")
        self.gridLayout_2.addWidget(self.label_image_hfw, 5, 0, 1, 1)
        self.label_detector_mode = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.label_detector_mode.setObjectName("label_detector_mode")
        self.gridLayout_2.addWidget(self.label_detector_mode, 28, 0, 1, 1)
        self.label_image_label = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.label_image_label.setObjectName("label_image_label")
        self.gridLayout_2.addWidget(self.label_image_label, 11, 0, 1, 1)
        self.checkBox_advanced_settings = QtWidgets.QCheckBox(self.scrollAreaWidgetContents)
        self.checkBox_advanced_settings.setObjectName("checkBox_advanced_settings")
        self.gridLayout_2.addWidget(self.checkBox_advanced_settings, 0, 0, 1, 2)
        self.label_beam_header = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        font = QtGui.QFont()
        font.setPointSize(10)
        font.setBold(True)
        font.setWeight(75)
        self.label_beam_header.setFont(font)
        self.label_beam_header.setObjectName("label_beam_header")
        self.gridLayout_2.addWidget(self.label_beam_header, 18, 0, 1, 3)
        self.stigmation_y = QtWidgets.QDoubleSpinBox(self.scrollAreaWidgetContents)
        self.stigmation_y.setMaximum(1000.0)
        self.stigmation_y.setObjectName("stigmation_y")
        self.gridLayout_2.addWidget(self.stigmation_y, 22, 2, 1, 1)
        self.spinBox_resolution_x = QtWidgets.QSpinBox(self.scrollAreaWidgetContents)
        self.spinBox_resolution_x.setMinimum(1)
        self.spinBox_resolution_x.setMaximum(999999999)
        self.spinBox_resolution_x.setProperty("value", 1536)
        self.spinBox_resolution_x.setObjectName("spinBox_resolution_x")
        self.gridLayout_2.addWidget(self.spinBox_resolution_x, 3, 1, 1, 1)
        self.label_beam_working_distance = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.label_beam_working_distance.setObjectName("label_beam_working_distance")
        self.gridLayout_2.addWidget(self.label_beam_working_distance, 19, 0, 1, 1)
        self.scalebar_checkbox = QtWidgets.QCheckBox(self.scrollAreaWidgetContents)
        self.scalebar_checkbox.setObjectName("scalebar_checkbox")
        self.gridLayout_2.addWidget(self.scalebar_checkbox, 34, 0, 1, 1)
        self.checkBox_image_use_autocontrast = QtWidgets.QCheckBox(self.scrollAreaWidgetContents)
        self.checkBox_image_use_autocontrast.setChecked(True)
        self.checkBox_image_use_autocontrast.setObjectName("checkBox_image_use_autocontrast")
        self.gridLayout_2.addWidget(self.checkBox_image_use_autocontrast, 7, 0, 1, 1)
        self.label_beam_current = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.label_beam_current.setObjectName("label_beam_current")
        self.gridLayout_2.addWidget(self.label_beam_current, 20, 0, 1, 1)
        self.ion_ruler_checkBox = QtWidgets.QCheckBox(self.scrollAreaWidgetContents)
        self.ion_ruler_checkBox.setObjectName("ion_ruler_checkBox")
        self.gridLayout_2.addWidget(self.ion_ruler_checkBox, 34, 2, 1, 1)
        self.spinBox_beam_scan_rotation = QtWidgets.QSpinBox(self.scrollAreaWidgetContents)
        self.spinBox_beam_scan_rotation.setMaximum(360)
        self.spinBox_beam_scan_rotation.setObjectName("spinBox_beam_scan_rotation")
        self.gridLayout_2.addWidget(self.spinBox_beam_scan_rotation, 24, 1, 1, 2)
        self.doubleSpinBox_image_dwell_time = QtWidgets.QDoubleSpinBox(self.scrollAreaWidgetContents)
        self.doubleSpinBox_image_dwell_time.setMaximum(10000000000000.0)
        self.doubleSpinBox_image_dwell_time.setSingleStep(0.01)
        self.doubleSpinBox_image_dwell_time.setProperty("value", 1.0)
        self.doubleSpinBox_image_dwell_time.setObjectName("doubleSpinBox_image_dwell_time")
        self.gridLayout_2.addWidget(self.doubleSpinBox_image_dwell_time, 4, 1, 1, 2)
        self.button_set_beam_settings = QtWidgets.QPushButton(self.scrollAreaWidgetContents)
        self.button_set_beam_settings.setObjectName("button_set_beam_settings")
        self.gridLayout_2.addWidget(self.button_set_beam_settings, 25, 0, 1, 3)
        self.selected_beam = QtWidgets.QComboBox(self.scrollAreaWidgetContents)
        self.selected_beam.setObjectName("selected_beam")
        self.gridLayout_2.addWidget(self.selected_beam, 1, 1, 1, 2)
        self.shift_y = QtWidgets.QDoubleSpinBox(self.scrollAreaWidgetContents)
        self.shift_y.setMinimum(-1000.0)
        self.shift_y.setMaximum(1000.0)
        self.shift_y.setObjectName("shift_y")
        self.gridLayout_2.addWidget(self.shift_y, 23, 2, 1, 1)
        self.comboBox_presets = QtWidgets.QComboBox(self.scrollAreaWidgetContents)
        self.comboBox_presets.setObjectName("comboBox_presets")
        self.gridLayout_2.addWidget(self.comboBox_presets, 6, 1, 1, 2)
        self.label_stigmation = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.label_stigmation.setObjectName("label_stigmation")
        self.gridLayout_2.addWidget(self.label_stigmation, 22, 0, 1, 1)
        self.detector_type_combobox = QtWidgets.QComboBox(self.scrollAreaWidgetContents)
        self.detector_type_combobox.setObjectName("detector_type_combobox")
        self.gridLayout_2.addWidget(self.detector_type_combobox, 27, 1, 1, 2)
        self.label_image_dwell_time = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.label_image_dwell_time.setObjectName("label_image_dwell_time")
        self.gridLayout_2.addWidget(self.label_image_dwell_time, 4, 0, 1, 1)
        self.label_presets = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.label_presets.setObjectName("label_presets")
        self.gridLayout_2.addWidget(self.label_presets, 6, 0, 1, 1)
        self.label_beam_scan_rotation = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.label_beam_scan_rotation.setObjectName("label_beam_scan_rotation")
        self.gridLayout_2.addWidget(self.label_beam_scan_rotation, 24, 0, 1, 1)
        self.label_detector_brightness = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.label_detector_brightness.setObjectName("label_detector_brightness")
        self.gridLayout_2.addWidget(self.label_detector_brightness, 29, 0, 1, 1)
        self.checkBox_image_use_autogamma = QtWidgets.QCheckBox(self.scrollAreaWidgetContents)
        self.checkBox_image_use_autogamma.setObjectName("checkBox_image_use_autogamma")
        self.gridLayout_2.addWidget(self.checkBox_image_use_autogamma, 7, 1, 1, 1)
        self.lineEdit_image_path = QtWidgets.QLineEdit(self.scrollAreaWidgetContents)
        self.lineEdit_image_path.setObjectName("lineEdit_image_path")
        self.gridLayout_2.addWidget(self.lineEdit_image_path, 10, 1, 1, 2)
        self.detector_brightness_label = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.detector_brightness_label.setObjectName("detector_brightness_label")
        self.gridLayout_2.addWidget(self.detector_brightness_label, 29, 2, 1, 1)
        self.spinBox_resolution_y = QtWidgets.QSpinBox(self.scrollAreaWidgetContents)
        self.spinBox_resolution_y.setMinimum(1)
        self.spinBox_resolution_y.setMaximum(999999999)
        self.spinBox_resolution_y.setProperty("value", 1024)
        self.spinBox_resolution_y.setObjectName("spinBox_resolution_y")
        self.gridLayout_2.addWidget(self.spinBox_resolution_y, 3, 2, 1, 1)
        self.label_image_save_path = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.label_image_save_path.setObjectName("label_image_save_path")
        self.gridLayout_2.addWidget(self.label_image_save_path, 10, 0, 1, 1)
        self.lineEdit_image_label = QtWidgets.QLineEdit(self.scrollAreaWidgetContents)
        self.lineEdit_image_label.setObjectName("lineEdit_image_label")
        self.gridLayout_2.addWidget(self.lineEdit_image_label, 11, 1, 1, 2)
        self.detector_mode_combobox = QtWidgets.QComboBox(self.scrollAreaWidgetContents)
        self.detector_mode_combobox.setObjectName("detector_mode_combobox")
        self.gridLayout_2.addWidget(self.detector_mode_combobox, 28, 1, 1, 2)
        self.label_image_resolution = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.label_image_resolution.setObjectName("label_image_resolution")
        self.gridLayout_2.addWidget(self.label_image_resolution, 3, 0, 1, 1)
        self.stigmation_x = QtWidgets.QDoubleSpinBox(self.scrollAreaWidgetContents)
        self.stigmation_x.setMinimum(-1000.0)
        self.stigmation_x.setMaximum(1000.0)
        self.stigmation_x.setObjectName("stigmation_x")
        self.gridLayout_2.addWidget(self.stigmation_x, 22, 1, 1, 1)
        self.detector_contrast_label = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.detector_contrast_label.setObjectName("detector_contrast_label")
        self.gridLayout_2.addWidget(self.detector_contrast_label, 30, 2, 1, 1)
        self.label_beam_voltage = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.label_beam_voltage.setObjectName("label_beam_voltage")
        self.gridLayout_2.addWidget(self.label_beam_voltage, 21, 0, 1, 1)
        self.shift_x = QtWidgets.QDoubleSpinBox(self.scrollAreaWidgetContents)
        self.shift_x.setMinimum(-1000.0)
        self.shift_x.setMaximum(1000.0)
        self.shift_x.setObjectName("shift_x")
        self.gridLayout_2.addWidget(self.shift_x, 23, 1, 1, 1)
        self.label_beam_type = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.label_beam_type.setObjectName("label_beam_type")
        self.gridLayout_2.addWidget(self.label_beam_type, 1, 0, 1, 1)
        self.ion_ruler_label = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.ion_ruler_label.setText("")
        self.ion_ruler_label.setObjectName("ion_ruler_label")
        self.gridLayout_2.addWidget(self.ion_ruler_label, 33, 1, 1, 2)
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.gridLayout_2.addItem(spacerItem, 35, 0, 1, 2)
        self.working_distance = QtWidgets.QDoubleSpinBox(self.scrollAreaWidgetContents)
        self.working_distance.setMaximum(1000.0)
        self.working_distance.setObjectName("working_distance")
        self.gridLayout_2.addWidget(self.working_distance, 19, 1, 1, 2)
        self.beam_current = QtWidgets.QDoubleSpinBox(self.scrollAreaWidgetContents)
        self.beam_current.setMaximum(1000.0)
        self.beam_current.setObjectName("beam_current")
        self.gridLayout_2.addWidget(self.beam_current, 20, 1, 1, 2)
        self.beam_voltage = QtWidgets.QDoubleSpinBox(self.scrollAreaWidgetContents)
        self.beam_voltage.setMaximum(100000.0)
        self.beam_voltage.setObjectName("beam_voltage")
        self.gridLayout_2.addWidget(self.beam_voltage, 21, 1, 1, 2)
        self.scrollArea.setWidget(self.scrollAreaWidgetContents)
        self.gridLayout.addWidget(self.scrollArea, 1, 0, 1, 2)
        self.pushButton_live_imaging = QtWidgets.QPushButton(Form)
        self.pushButton_live_imaging.setObjectName("pushButton_live_imaging")
        self.gridLayout.addWidget(self.pushButton_live_imaging, 31, 0, 1, 2)

        self.retranslateUi(Form)
        QtCore.QMetaObject.connectSlotsByName(Form)
        Form.setTabOrder(self.pushButton_take_image, self.pushButton_take_all_images)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Form"))
        self.pushButton_take_image.setText(_translate("Form", "Acquire Image"))
        self.pushButton_take_all_images.setText(_translate("Form", "Acquire All Images"))
        self.label_detector_header.setText(_translate("Form", "Detector Settings"))
        self.label_detector_contrast.setText(_translate("Form", "Contrast (%)"))
        self.label_shift.setText(_translate("Form", "Shift (x,y) (um)"))
        self.label_title.setText(_translate("Form", "Image Settings"))
        self.label_utils_header.setText(_translate("Form", "Utilities"))
        self.crosshair_checkbox.setText(_translate("Form", "Cross Hair"))
        self.checkBox_image_save_image.setText(_translate("Form", "Save Image"))
        self.label_detector_type.setText(_translate("Form", "Type"))
        self.set_detector_button.setText(_translate("Form", "Set Detector Settings"))
        self.label_image_hfw.setText(_translate("Form", "HFW (um)"))
        self.label_detector_mode.setText(_translate("Form", "Mode"))
        self.label_image_label.setText(_translate("Form", "Filename"))
        self.checkBox_advanced_settings.setText(_translate("Form", "Advanced Settings"))
        self.label_beam_header.setText(_translate("Form", "Beam Settings"))
        self.label_beam_working_distance.setText(_translate("Form", "Working Distance (mm)"))
        self.scalebar_checkbox.setText(_translate("Form", "Scale Bar"))
        self.checkBox_image_use_autocontrast.setText(_translate("Form", "AutoContrast"))
        self.label_beam_current.setText(_translate("Form", "Beam Current (pA)"))
        self.ion_ruler_checkBox.setText(_translate("Form", "Ruler "))
        self.button_set_beam_settings.setText(_translate("Form", "Set Beam Settings"))
        self.label_stigmation.setText(_translate("Form", "Stigmation (x,y)"))
        self.label_image_dwell_time.setText(_translate("Form", "Dwell Time (us)"))
        self.label_presets.setText(_translate("Form", "Preset"))
        self.label_beam_scan_rotation.setText(_translate("Form", "Scan Rotation (deg)"))
        self.label_detector_brightness.setText(_translate("Form", "Brightness (%)"))
        self.checkBox_image_use_autogamma.setText(_translate("Form", "AutoGamma"))
        self.detector_brightness_label.setText(_translate("Form", "50%"))
        self.label_image_save_path.setText(_translate("Form", "Path"))
        self.label_image_resolution.setText(_translate("Form", "Resolution (px)"))
        self.detector_contrast_label.setText(_translate("Form", "50%"))
        self.label_beam_voltage.setText(_translate("Form", "Voltage (kV)"))
        self.label_beam_type.setText(_translate("Form", "Beam"))
        self.pushButton_live_imaging.setText(_translate("Form", "Live Imaging"))




import glob
import os
from copy import deepcopy
from pathlib import Path
from pprint import pprint

import matplotlib.pyplot as plt
import pandas as pd
from tqdm import tqdm

from fibsem.detection import detection
from fibsem.segmentation.model import load_model
from fibsem.segmentation.utils import decode_segmap_v2
from fibsem.structures import FibsemImage, Point

import tifffile as tff


def _run_evaluation(path:Path, image_path: Path, checkpoints: list[dict], plot: bool = False, _FEATURES_TO_IGNORE: list[str] = ["ImageCentre", "LandingPost"], save: bool = False, save_path: Path = None):

    # ground truth 
    df = pd.read_csv(path)

    filename_col = "image" if "image" in df.columns else "filename"
    image_fnames = df[filename_col].unique().tolist()

    # glob each image in the image_path and add to filenames
    filenames = []
    
    for fname in image_fnames:
        fname = fname.removesuffix(".tif")
        filenames += glob.glob(os.path.join(image_path, f"{fname}*.tif"))
    
    print(f"Found {len(filenames)} images (test)")

    # setup eval dataframe
    px_col = "p.x" if "p.x" in df.columns else "px.x"
    py_col = "p.y" if "p.y" in df.columns else "px.y"
    df_eval = pd.DataFrame(columns=["checkpoint", "fname", "feature", px_col, py_col, "pixelsize", "gt.p.x", "gt.p.y", "gt.pixelsize", "distance","_is_corrected"])
    df_eval["_is_corrected"] = df_eval["_is_corrected"].astype(bool)

    _prog = tqdm(sorted(filenames))
    for i, fname in enumerate(_prog):
        
        image_fname = os.path.basename(fname).removesuffix(".tif")

        # if suffix is either _eb or _ib, remove it
        if image_fname.endswith("_eb") or image_fname.endswith("_ib"):
            image_fname = image_fname[:-3]

        df_filt = df[df[filename_col] == image_fname]

        if len(df_filt) == 0:
            print(f"no ground truth for {fname}")
            continue

        _is_corrected = df_filt["corrected"].values[0]

        # extract each row
        gt_features = []
        pixelsize = 25e-9
        for _, row in df_filt.iterrows():

            feature = detection.get_feature(row["feature"])
            feature.px = Point(row[px_col], row[py_col])
            pixelsize= row["pixelsize"]
            
            if feature.name in _FEATURES_TO_IGNORE:
                continue

            gt_features.append(feature)
        

        image = FibsemImage.load(fname)

        # plot
        # TODO: updat when get gt mask / rgb
        gt_det = detection.DetectedFeatures(
            features = gt_features,
            image=image.data,
            mask=None,
            rgb=None,
            pixelsize=pixelsize, 
            fibsem_image=image)

        dets = [gt_det]
        for k, CKPT in enumerate(checkpoints):

            CHECKPOINT = CKPT["checkpoint"]
            ENCODER = CKPT["encoder"]
            NC = CKPT["nc"]
            

            _prog.set_description(f"file ({i+1}/{len(filenames)}) {os.path.basename(fname)} - checkpoint: ({k+1}/{len(checkpoints)}) {os.path.basename(CHECKPOINT)}")
            model = load_model(CHECKPOINT, encoder=ENCODER, nc=NC)

            det  = detection.detect_features(image.data, model, features=deepcopy(gt_features), pixelsize=pixelsize, filter=True)
            dets.append(det)
        
        if plot or save:
            fig = detection.plot_detections(dets, titles=["Ground Truth"] + [os.path.basename(ckpt["checkpoint"].removesuffix(".pt")) for ckpt in checkpoints])
            if plot:
                plt.show()

            if save:
                os.makedirs(save_path, exist_ok=True)
                fig.savefig(os.path.join(save_path, f"{image_fname}.png"), dpi=300)
                plt.close(fig)

        # evaluation against ground truth
        for ckpt, det in zip(checkpoints, dets[1:]):
            checkpoint = ckpt["checkpoint"]
            for feature in det.features:
                gt_feat = gt_det.get_feature(feature.name)
                dpx = feature.px - gt_feat.px
                    
                df_eval2 = pd.DataFrame([{
                    "checkpoint": os.path.basename(checkpoint),
                    "fname": os.path.basename(fname),
                    "gt_fname": image_fname,
                    "feature": feature.name,
                    "p.x": feature.px.x,
                    "p.y": feature.px.y,
                    "pixelsize": pixelsize,
                    "gt.p.x": gt_feat.px.x,
                    "gt.p.y": gt_feat.px.y,
                    "gt.pixelsize": pixelsize,
                    "distance": feature.px.euclidean(gt_feat.px),
                    "d.p.x": dpx.x,
                    "d.p.y": dpx.y,
                    "_is_corrected": bool(_is_corrected), 
                }])

                # convert to bool type
                df_eval2["_is_corrected"] = df_eval2["_is_corrected"].astype(bool)
                df_eval = pd.concat([df_eval, df_eval2], axis=0, ignore_index=True)
        
        # print("-"*80, "\n")
        # if i==3:
            # break

    df_eval["checkpoint"] = df_eval["checkpoint"].str.replace(".pt", "")

    if save:
        df_eval.to_csv(os.path.join(save_path, "eval.csv"), index=False)

    return df_eval



def run_evaluation_v2(path:Path, image_path: Path, checkpoints: list[dict], labels_path: Path = None, 
                      plot: bool = False, _FEATURES_TO_IGNORE: list[str] = ["ImageCentre"], 
                      save: bool = False, save_path: Path = None):

    # ground truth 
    df = pd.read_csv(path)

    image_fnames = df["filename"].unique().tolist()

    # glob each image in the image_path and add to filenames
    filenames = []
    
    # TODO: keep track of whether .tif ext was recorded in the csv

    for fname in image_fnames:
        fname = fname.removesuffix(".tif")
        filenames += glob.glob(os.path.join(image_path, f"{fname}*.tif"))
    
    print(f"Found {len(filenames)} images (test)")

    if labels_path is None:
        labels_path = os.path.join(image_path, "labels")

    # setup eval dataframe
    df_eval = pd.DataFrame(columns=["checkpoint", "fname", "feature", "px.x", "px.y", "pixelsize", "gt.p.x", "gt.p.y", "gt.pixelsize", "distance","_is_corrected"])
    df_eval["_is_corrected"] = df_eval["_is_corrected"].astype(bool)

    _prog = tqdm(sorted(filenames))
    for i, fname in enumerate(_prog):
        
        image_fname = os.path.basename(fname)#.removesuffix(".tif")

        # if suffix is either _eb or _ib, remove it
        if image_fname.endswith("_eb") or image_fname.endswith("_ib"):
            image_fname = image_fname[:-3]

        df_filt = df[df["filename"] == image_fname]

        if len(df_filt) == 0:
            print(f"no ground truth for {fname}")
            continue

        _is_corrected = df_filt["corrected"].values[0]

        # extract each row
        gt_features = []
        pixelsize = 25e-9
        for _, row in df_filt.iterrows():

            # if entries in row are null, go to next image
            if pd.isnull(row["feature"]):
                continue

            feature = detection.get_feature(row["feature"])
            feature.px = Point(row["px.x"], row["px.y"])
            pixelsize= row["pixelsize"]
            
            if feature.name in _FEATURES_TO_IGNORE:
                continue

            gt_features.append(feature)
        

        image = FibsemImage.load(fname)

        # plot
        mask, rgb = None, None
        label_fname = os.path.join(labels_path, f"{image_fname}")
        if os.path.exists(label_fname):
            mask = tff.imread(os.path.join(labels_path, f"{image_fname}"))
            rgb = decode_segmap_v2(mask)
        gt_det = detection.DetectedFeatures(
            features = gt_features,
            image=image.data,
            mask=mask,
            rgb=rgb,
            pixelsize=pixelsize, 
            fibsem_image=image)

        dets = [gt_det]
        for k, CKPT in enumerate(checkpoints):

            CHECKPOINT = CKPT["checkpoint"]           

            _prog.set_description(f"file ({i+1}/{len(filenames)}) {os.path.basename(fname)} - checkpoint: ({k+1}/{len(checkpoints)}) {os.path.basename(CHECKPOINT)}")
            model = load_model(CHECKPOINT) 

            det  = detection.detect_features(image.data, model, features=deepcopy(gt_features), pixelsize=pixelsize, filter=True)
            dets.append(det)
        
        if plot or save:
            fig = detection.plot_detections(dets, titles=["Ground Truth"] + [os.path.basename(ckpt["checkpoint"].removesuffix(".pt")) for ckpt in checkpoints])
            if plot:
                plt.show()

            if save:
                os.makedirs(save_path, exist_ok=True)
                fig.savefig(os.path.join(save_path, f"{image_fname}.png"), dpi=300)
                plt.close(fig)

        # evaluation against ground truth
        for ckpt, det in zip(checkpoints, dets[1:]):
            checkpoint = ckpt["checkpoint"]
            for feature in det.features:
                gt_feat = gt_det.get_feature(feature.name)
                dpx = feature.px - gt_feat.px
                    
                df_eval2 = pd.DataFrame([{
                    "checkpoint": os.path.basename(checkpoint),
                    "fname": os.path.basename(fname),
                    "gt_fname": image_fname,
                    "feature": feature.name,
                    "p.x": feature.px.x,
                    "p.y": feature.px.y,
                    "pixelsize": pixelsize,
                    "gt.p.x": gt_feat.px.x,
                    "gt.p.y": gt_feat.px.y,
                    "gt.pixelsize": pixelsize,
                    "distance": feature.px.euclidean(gt_feat.px),
                    "d.p.x": dpx.x,
                    "d.p.y": dpx.y,
                    "_is_corrected": bool(_is_corrected), 
                }])

                # convert to bool type
                df_eval2["_is_corrected"] = df_eval2["_is_corrected"].astype(bool)
                df_eval = pd.concat([df_eval, df_eval2], axis=0, ignore_index=True)
        
        # print("-"*80, "\n")
        # if i==3:
            # break

    df_eval["checkpoint"] = df_eval["checkpoint"].str.replace(".pt", "")

    if save:
        df_eval.to_csv(os.path.join(save_path, "eval.csv"), index=False)

    return df_eval


import plotly.express as px


def plot_evaluation_data(df: pd.DataFrame, category_orders: dict, show: bool = True, 
                            thresholds: list[int] =  [50, 25, 10], save: bool=False, save_path: Path = None):
    # scatter plot
    fig = px.scatter(df, x="d.p.x", y="d.p.y", color="feature", facet_col="checkpoint", 
        # facet_row="dataset", 
        category_orders=category_orders,
        title="Distance from Ground Truth", hover_data=df.columns)
    
    if save:
        # save fig as png
        fig.write_image(os.path.join(save_path, "eval00.png"))
    
    if show:
        fig.show()


    # plot histogram
    fig = px.histogram(df, x="distance", color="feature", facet_col="checkpoint", nbins=100, title="Distance from Ground Truth", hover_data=df.columns, 
                    category_orders=category_orders)
    if show:
        fig.show()

    if save:
        # save fig as png
        fig.write_image(os.path.join(save_path, "eval01.png"))
    

    # plot box plot
    fig = px.box(df, x="feature", y="distance", color="feature", facet_col="checkpoint", title="Distance from Ground Truth", hover_data=df.columns,
                    category_orders=category_orders)

    if show:
        fig.show()
    
    if save:
        # save fig as png
        fig.write_image(os.path.join(save_path, "eval02.png"))
    
    # # calculate the mean and standard deviation in one aggregation per checkpoint, feature
    df_agg = df.groupby(["checkpoint", "feature"]).agg({"distance": ["mean", "median", "std"]}).reset_index()

    # display(df_agg)
    
    # plot percentage of distance under threshold
    df_ret = pd.DataFrame()
    for threshold in thresholds:

        df[f"under_threshold_{threshold}px"] = df["distance"] < threshold

        df_group = df.groupby(["checkpoint", "feature", f"under_threshold_{threshold}px"]).count().reset_index()
        # # pivot table
        df_group = df_group.pivot(index=["checkpoint", "feature"], columns=f"under_threshold_{threshold}px", values="distance").reset_index()
        
        df_group["threshold"] = threshold

        # add false if not present
        if False not in df_group.columns:
            df_group[False] = 0

        # add true if not present
        if True not in df_group.columns:
            df_group[True] = 0

        # fill na with 0
        df_group.fillna(0, inplace=True)

        # # calc percentage
        df_group[f"under_threshold"] = df_group[True] / (df_group[True] + df_group[False])
        
        # concat to df_ret
        df_ret = pd.concat([df_ret, df_group], axis=0)


    df_ret = df_ret[["checkpoint", "feature", "threshold", "under_threshold"]]

    # plot on bar chart with plotly express

    # sort column by category order
    df_ret["feature"] = pd.Categorical(df_ret["feature"], category_orders["feature"])
    df_ret["checkpoint"] = pd.Categorical(df_ret["checkpoint"], category_orders["checkpoint"])
    df_ret.sort_values(["checkpoint", "feature"], inplace=True)
    df_ret["threshold"] = df_ret["threshold"].astype(str)

    fig = px.bar(df_ret, x="threshold", y="under_threshold", color="feature", facet_col="checkpoint", barmode="group",                    
                        text=df_ret[f"under_threshold"].apply(lambda x: f"{x:.2f}"),
                        category_orders=category_orders, 
                        hover_data=df_ret.columns,
                        title="Percentage of Distance under threshold (px)")
    
    if show:
        fig.show()
    
    if save:
        # save fig as png
        fig.write_image(os.path.join(save_path, "eval03.png"))
    

    df_ret["threshold"] = df_ret["threshold"].astype(int)

    fig = px.line(df_ret, x="threshold", y="under_threshold", 
                  color="feature",
                    facet_col="checkpoint", 
                    # #facet_col="checkpoint", 
                    # text=df_ret[f"under_threshold"].apply(lambda x: f"{x:.2f}"),
                    category_orders=category_orders,
                    title="Percentage of Distance under threshold (px)")
    
    # reverse order of x axis
    fig.update_xaxes(autorange="reversed")
    if show:
        fig.show()

    if save:
        # save fig as png
        fig.write_image(os.path.join(save_path, "eval04.png"))
    

    # display(df_ret)
        
    # calculate the accuracy at each threshold, save to csv
    df_ret["accuracy"] = df_ret["under_threshold"] * 100
    df_ret["accuracy"] = df_ret["accuracy"]
    df_ret["threshold"] = df_ret["threshold"].astype(int)

    df_ret.to_csv(os.path.join(save_path, "eval_accuracy.csv"), index=False)


    # return 



# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'FibsemModelTrainingWidget.ui'
#
# Created by: PyQt5 UI code generator 5.15.9
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_Form(object):
    def setupUi(self, Form):
        Form.setObjectName("Form")
        Form.resize(872, 888)
        self.gridLayout = QtWidgets.QGridLayout(Form)
        self.gridLayout.setObjectName("gridLayout")
        self.scrollArea = QtWidgets.QScrollArea(Form)
        self.scrollArea.setWidgetResizable(True)
        self.scrollArea.setObjectName("scrollArea")
        self.scrollAreaWidgetContents = QtWidgets.QWidget()
        self.scrollAreaWidgetContents.setGeometry(QtCore.QRect(0, 0, 838, 478))
        self.scrollAreaWidgetContents.setObjectName("scrollAreaWidgetContents")
        self.gridLayout_6 = QtWidgets.QGridLayout(self.scrollAreaWidgetContents)
        self.gridLayout_6.setObjectName("gridLayout_6")
        self.spinBox_batch_size = QtWidgets.QSpinBox(self.scrollAreaWidgetContents)
        self.spinBox_batch_size.setMaximum(256)
        self.spinBox_batch_size.setProperty("value", 4)
        self.spinBox_batch_size.setObjectName("spinBox_batch_size")
        self.gridLayout_6.addWidget(self.spinBox_batch_size, 10, 1, 1, 2)
        self.label_checkpoint = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.label_checkpoint.setObjectName("label_checkpoint")
        self.gridLayout_6.addWidget(self.label_checkpoint, 7, 0, 1, 1)
        self.lineEdit_save_path = QtWidgets.QLineEdit(self.scrollAreaWidgetContents)
        self.lineEdit_save_path.setObjectName("lineEdit_save_path")
        self.gridLayout_6.addWidget(self.lineEdit_save_path, 4, 1, 1, 1)
        self.label_epochs = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.label_epochs.setObjectName("label_epochs")
        self.gridLayout_6.addWidget(self.label_epochs, 11, 0, 1, 1)
        self.doubleSpinBox_lr = QtWidgets.QDoubleSpinBox(self.scrollAreaWidgetContents)
        self.doubleSpinBox_lr.setDecimals(4)
        self.doubleSpinBox_lr.setSingleStep(0.0001)
        self.doubleSpinBox_lr.setProperty("value", 0.0001)
        self.doubleSpinBox_lr.setObjectName("doubleSpinBox_lr")
        self.gridLayout_6.addWidget(self.doubleSpinBox_lr, 12, 1, 1, 2)
        self.lineEdit_data_path = QtWidgets.QLineEdit(self.scrollAreaWidgetContents)
        self.lineEdit_data_path.setObjectName("lineEdit_data_path")
        self.gridLayout_6.addWidget(self.lineEdit_data_path, 2, 1, 1, 1)
        self.labelPath_button = QtWidgets.QToolButton(self.scrollAreaWidgetContents)
        self.labelPath_button.setObjectName("labelPath_button")
        self.gridLayout_6.addWidget(self.labelPath_button, 3, 2, 1, 1)
        self.label_num_classes = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.label_num_classes.setObjectName("label_num_classes")
        self.gridLayout_6.addWidget(self.label_num_classes, 8, 0, 1, 1)
        self.comboBox_encoder = QtWidgets.QComboBox(self.scrollAreaWidgetContents)
        self.comboBox_encoder.setObjectName("comboBox_encoder")
        self.gridLayout_6.addWidget(self.comboBox_encoder, 6, 1, 1, 2)
        self.lineEdit_checkpoint = QtWidgets.QLineEdit(self.scrollAreaWidgetContents)
        self.lineEdit_checkpoint.setObjectName("lineEdit_checkpoint")
        self.gridLayout_6.addWidget(self.lineEdit_checkpoint, 7, 1, 1, 2)
        self.label_wandb_entity = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.label_wandb_entity.setObjectName("label_wandb_entity")
        self.gridLayout_6.addWidget(self.label_wandb_entity, 15, 0, 1, 1)
        self.spinBox_num_classes = QtWidgets.QSpinBox(self.scrollAreaWidgetContents)
        self.spinBox_num_classes.setObjectName("spinBox_num_classes")
        self.gridLayout_6.addWidget(self.spinBox_num_classes, 8, 1, 1, 2)
        self.label_encoder = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.label_encoder.setObjectName("label_encoder")
        self.gridLayout_6.addWidget(self.label_encoder, 6, 0, 1, 1)
        self.checkBox_use_wandb = QtWidgets.QCheckBox(self.scrollAreaWidgetContents)
        self.checkBox_use_wandb.setObjectName("checkBox_use_wandb")
        self.gridLayout_6.addWidget(self.checkBox_use_wandb, 14, 0, 1, 1)
        self.label_label_path = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.label_label_path.setObjectName("label_label_path")
        self.gridLayout_6.addWidget(self.label_label_path, 3, 0, 1, 1)
        self.label_data_path = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.label_data_path.setObjectName("label_data_path")
        self.gridLayout_6.addWidget(self.label_data_path, 2, 0, 1, 1)
        self.outputPath_button = QtWidgets.QToolButton(self.scrollAreaWidgetContents)
        self.outputPath_button.setObjectName("outputPath_button")
        self.gridLayout_6.addWidget(self.outputPath_button, 4, 2, 1, 1)
        self.label_learning_rate = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.label_learning_rate.setObjectName("label_learning_rate")
        self.gridLayout_6.addWidget(self.label_learning_rate, 12, 0, 1, 1)
        self.lineEdit_label_path = QtWidgets.QLineEdit(self.scrollAreaWidgetContents)
        self.lineEdit_label_path.setObjectName("lineEdit_label_path")
        self.gridLayout_6.addWidget(self.lineEdit_label_path, 3, 1, 1, 1)
        self.label = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.label.setObjectName("label")
        self.gridLayout_6.addWidget(self.label, 4, 0, 1, 1)
        self.label_batch_size = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.label_batch_size.setObjectName("label_batch_size")
        self.gridLayout_6.addWidget(self.label_batch_size, 10, 0, 1, 1)
        self.spinBox_epochs = QtWidgets.QSpinBox(self.scrollAreaWidgetContents)
        self.spinBox_epochs.setMaximum(9999)
        self.spinBox_epochs.setProperty("value", 20)
        self.spinBox_epochs.setObjectName("spinBox_epochs")
        self.gridLayout_6.addWidget(self.spinBox_epochs, 11, 1, 1, 2)
        self.dataPath_button = QtWidgets.QToolButton(self.scrollAreaWidgetContents)
        self.dataPath_button.setObjectName("dataPath_button")
        self.gridLayout_6.addWidget(self.dataPath_button, 2, 2, 1, 1)
        self.label_wandb_project = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        self.label_wandb_project.setObjectName("label_wandb_project")
        self.gridLayout_6.addWidget(self.label_wandb_project, 16, 0, 1, 1)
        self.lineEdit_wandb_entity = QtWidgets.QLineEdit(self.scrollAreaWidgetContents)
        self.lineEdit_wandb_entity.setObjectName("lineEdit_wandb_entity")
        self.gridLayout_6.addWidget(self.lineEdit_wandb_entity, 15, 1, 1, 1)
        self.lineEdit_wandb_project = QtWidgets.QLineEdit(self.scrollAreaWidgetContents)
        self.lineEdit_wandb_project.setContextMenuPolicy(QtCore.Qt.DefaultContextMenu)
        self.lineEdit_wandb_project.setObjectName("lineEdit_wandb_project")
        self.gridLayout_6.addWidget(self.lineEdit_wandb_project, 16, 1, 1, 1)
        self.label_2 = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.label_2.setFont(font)
        self.label_2.setObjectName("label_2")
        self.gridLayout_6.addWidget(self.label_2, 0, 0, 1, 3)
        self.label_3 = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.label_3.setFont(font)
        self.label_3.setObjectName("label_3")
        self.gridLayout_6.addWidget(self.label_3, 5, 0, 1, 3)
        self.label_4 = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.label_4.setFont(font)
        self.label_4.setObjectName("label_4")
        self.gridLayout_6.addWidget(self.label_4, 9, 0, 1, 3)
        self.label_5 = QtWidgets.QLabel(self.scrollAreaWidgetContents)
        font = QtGui.QFont()
        font.setBold(True)
        font.setWeight(75)
        self.label_5.setFont(font)
        self.label_5.setObjectName("label_5")
        self.gridLayout_6.addWidget(self.label_5, 13, 0, 1, 3)
        self.scrollArea.setWidget(self.scrollAreaWidgetContents)
        self.gridLayout.addWidget(self.scrollArea, 1, 0, 1, 3)
        self.label_info = QtWidgets.QLabel(Form)
        self.label_info.setObjectName("label_info")
        self.gridLayout.addWidget(self.label_info, 26, 0, 1, 2)
        self.pushButton_train_model = QtWidgets.QPushButton(Form)
        self.pushButton_train_model.setObjectName("pushButton_train_model")
        self.gridLayout.addWidget(self.pushButton_train_model, 25, 0, 1, 3)
        spacerItem = QtWidgets.QSpacerItem(20, 40, QtWidgets.QSizePolicy.Minimum, QtWidgets.QSizePolicy.Expanding)
        self.gridLayout.addItem(spacerItem, 2, 0, 1, 3)
        self.label_title = QtWidgets.QLabel(Form)
        font = QtGui.QFont()
        font.setPointSize(12)
        font.setBold(True)
        font.setWeight(75)
        self.label_title.setFont(font)
        self.label_title.setObjectName("label_title")
        self.gridLayout.addWidget(self.label_title, 0, 0, 1, 3)

        self.retranslateUi(Form)
        QtCore.QMetaObject.connectSlotsByName(Form)

    def retranslateUi(self, Form):
        _translate = QtCore.QCoreApplication.translate
        Form.setWindowTitle(_translate("Form", "Form"))
        self.label_checkpoint.setText(_translate("Form", "Checkpoint"))
        self.label_epochs.setText(_translate("Form", "Epochs"))
        self.labelPath_button.setText(_translate("Form", "..."))
        self.label_num_classes.setText(_translate("Form", "Num Classes"))
        self.label_wandb_entity.setText(_translate("Form", "Entity"))
        self.label_encoder.setText(_translate("Form", "Encoder"))
        self.checkBox_use_wandb.setText(_translate("Form", "Use Weights and Biases"))
        self.label_label_path.setText(_translate("Form", "Label Path"))
        self.label_data_path.setText(_translate("Form", "Data Path"))
        self.outputPath_button.setText(_translate("Form", "..."))
        self.label_learning_rate.setText(_translate("Form", "Learning Rate"))
        self.label.setText(_translate("Form", "Save Path"))
        self.label_batch_size.setText(_translate("Form", "Batch Size"))
        self.dataPath_button.setText(_translate("Form", "..."))
        self.label_wandb_project.setText(_translate("Form", "Project"))
        self.label_2.setText(_translate("Form", "Data"))
        self.label_3.setText(_translate("Form", "Model"))
        self.label_4.setText(_translate("Form", "Hyper Parameters"))
        self.label_5.setText(_translate("Form", "Logging"))
        self.label_info.setText(_translate("Form", "Info"))
        self.pushButton_train_model.setText(_translate("Form", "Train Model"))
        self.label_title.setText(_translate("Form", "Model Training"))



# style sheets
# TODO: TEMPLATE THIS CSS   
_GREEN_PUSHBUTTON_STYLE = """
QPushButton {
    background-color: green;
    }
QPushButton:hover {
    background-color: rgba(0, 255, 0, 125);
    }"""

_RED_PUSHBUTTON_STYLE = """
QPushButton {
    background-color: red;
    }
QPushButton:hover {
    background-color: rgba(255, 0, 0, 125);
    }"""

_BLUE_PUSHBUTTON_STYLE = """
QPushButton {
    background-color: blue;
    }
QPushButton:hover {
    background-color: rgba(0, 0, 255, 125);
    }"""

_YELLOW_PUSHBUTTON_STYLE = """
QPushButton {
    background-color: yellow;
    }
QPushButton:hover {
    background-color: rgba(255, 255, 0, 125);
    }"""

_WHITE_PUSHBUTTON_STYLE = """
QPushButton {
    background-color: white;
    }
QPushButton:hover {
    background-color: rgba(255, 255, 255, 125);
    }"""

_GRAY_PUSHBUTTON_STYLE = """
QPushButton {
    background-color: gray;
    }
QPushButton:hover {
    background-color: rgba(125, 125, 125, 125);
    }"""

_ORANGE_PUSHBUTTON_STYLE = """
QPushButton {
    background-color: orange;
    }
QPushButton:hover {
    background-color: rgba(255, 125, 0, 125);
}"""

_DISABLED_PUSHBUTTON_STYLE = """
QPushButton {
    background-color: none;
    }
"""    


import os
from copy import deepcopy

import napari
import napari.utils.notifications
import numpy as np
from PyQt5 import QtWidgets

from fibsem.detection import detection
from fibsem.detection import utils as det_utils
from fibsem.detection.detection import DetectedFeatures
from fibsem.segmentation import model as fibsem_model
from fibsem.segmentation.model import load_model
from fibsem.structures import (
    FibsemImage,
    Point,
)
from fibsem import utils
from fibsem.ui import _stylesheets
from PyQt5.QtCore import pyqtSignal
from fibsem.ui.qtdesigner_files import FibsemEmbeddedDetectionWidget
import logging

from fibsem.segmentation.config import CLASS_COLORS

class FibsemEmbeddedDetectionUI(FibsemEmbeddedDetectionWidget.Ui_Form, QtWidgets.QWidget):
    continue_signal = pyqtSignal(DetectedFeatures)

    def __init__(
        self,
        viewer: napari.Viewer,
        det: DetectedFeatures = None,
        model: fibsem_model.SegmentationModel = None,
        parent=None,
    ):
        super(FibsemEmbeddedDetectionUI, self).__init__(parent=parent)
        self.setupUi(self)

        self.parent = parent
        self.viewer = viewer
        self.viewer.window._qt_viewer.dockLayerList.setVisible(False)
        self.viewer.window._qt_viewer.dockLayerControls.setVisible(False)
        self.model = model

        self._USER_CORRECTED = False
        self._LABELLING_ENABLED = False
        self._MODEL_ASSIST = False

        self._image_layer = None
        self._mask_layer = None
        self._features_layer = None
        self._cross_hair_layer = None

        self.setup_connections()

        # set detected features
        if det is not None:
            self.set_detected_features(det)

        if model is not None:
            self._set_model(model)

    def setup_connections(self):
        self.label_instructions.setText(
            """Drag the detected feature positions to move them. Press Continue when finished."""
        )
        self.pushButton_continue.clicked.connect(self.confirm_button_clicked)
        self.pushButton_continue.setVisible(False)


        # labelling
        self.viewer.bind_key("L", self._toggle_labelling, overwrite=True)
        self.pushButton_enable_labelling.clicked.connect(self._toggle_labelling)
        self.checkBox_labelling_model_assist.stateChanged.connect(self._toggle_labelling)
        self.pushButton_enable_labelling.setStyleSheet(_stylesheets._GREEN_PUSHBUTTON_STYLE)  

        self.checkBox_labelling_model_assist.setVisible(False) # TODO: add model assist
        self.pushButton_labelling_confirm.setVisible(self._MODEL_ASSIST)
        self.pushButton_labelling_cancel.setVisible(self._MODEL_ASSIST)       
        self.label_labelling_class_index.setVisible(self._MODEL_ASSIST)
        self.label_labelling_instructions.setVisible(self._MODEL_ASSIST)
        self.label_labelling_model.setVisible(self._MODEL_ASSIST)
        self.comboBox_labelling_class_index.setVisible(self._MODEL_ASSIST)


    def _toggle_labelling(self, event=None):
        """Toggle labelling mode on/off"""
        if self.sender() != self.checkBox_labelling_model_assist:
            self._LABELLING_ENABLED = not self._LABELLING_ENABLED

        self.checkBox_labelling_model_assist.setVisible(False) # TODO: add model assist
        # self._MODEL_ASSIST = self.checkBox_labelling_model_assist.isChecked()
        self._MODEL_ASSIST = False

        self.pushButton_labelling_confirm.setVisible(self._MODEL_ASSIST)
        self.pushButton_labelling_cancel.setVisible(self._MODEL_ASSIST)       
        self.label_labelling_class_index.setVisible(self._MODEL_ASSIST)
        self.label_labelling_instructions.setVisible(self._MODEL_ASSIST)
        self.label_labelling_model.setVisible(self._MODEL_ASSIST)
        self.comboBox_labelling_class_index.setVisible(self._MODEL_ASSIST)

        # show layer controls
        self.viewer.window._qt_viewer.dockLayerList.setVisible(self._LABELLING_ENABLED)
        self.viewer.window._qt_viewer.dockLayerControls.setVisible(self._LABELLING_ENABLED)

        if self._LABELLING_ENABLED:
            self.viewer.layers.selection.active = self._mask_layer
            self._mask_layer.mode = "paint"
            self.pushButton_enable_labelling.setText("Disable Labelling")
            self.pushButton_enable_labelling.setStyleSheet(_stylesheets._ORANGE_PUSHBUTTON_STYLE)        
        else:
            self.viewer.layers.selection.active = self._features_layer
            self._features_layer.mode = "select"
            self.pushButton_enable_labelling.setText("Enable Labelling")
            self.pushButton_enable_labelling.setStyleSheet(_stylesheets._GREEN_PUSHBUTTON_STYLE)  

    def confirm_button_clicked(self, reset_camera=False):
        """Confirm the detected features, save the data and and remove the layers from the viewer."""
        
        # log the difference between initial and final detections
        try:
            fname = self.det.fibsem_image.metadata.image_settings.filename
            beam_type = self.det.fibsem_image.metadata.image_settings.beam_type
        except:
            fname = f"ml-{utils.current_timestamp_v2()}"
            beam_type = "NULL"
        
        fd = [] # feature detections
        for f0, f1 in zip(self.det.features, self._intial_det.features):
            px_diff = f1.px - f0.px
            msgd = {"msg": "feature_detection",
                    "fname": fname,                                             # filename
                    "feature": f0.name,                                         # feature name
                    "px": f0.px.to_dict(),                                      # pixel coordinates
                    "dpx": px_diff.to_dict(),                                   # pixel difference
                    "dm": px_diff._to_metres(self.det.pixelsize).to_dict(),     # metre difference
                    "is_correct": not np.any(px_diff),                          # is the feature correct    
                    "beam_type": beam_type.name,                                # beam type         
                    "pixelsize": self.det.pixelsize,                            # pixelsize
                    "checkpoint": self.det.checkpoint,                          # checkpoint
            }
            logging.debug(msgd)
            fd.append(deepcopy(msgd))                                           # to write to disk

        # save features data
        self.det.mask = self._mask_layer.data.astype(np.uint8) # type: ignore
        det_utils.save_feature_data_to_csv(self.det, features=fd, filename=fname)
            
        # remove feature detection layers
        if self._image_layer is not None:
            if self._image_layer in self.viewer.layers:
                self.viewer.layers.remove(self._image_layer)
            if self._mask_layer in self.viewer.layers:
                self.viewer.layers.remove(self._mask_layer)
            if self._features_layer in self.viewer.layers:
                self.viewer.layers.remove(self._features_layer)
            if self._cross_hair_layer in self.viewer.layers:
                self.viewer.layers.remove(self._cross_hair_layer)

        # reshow all other layers
        for layer in self.viewer.layers:
            layer.visible = True
        
        # reset camera
        self.viewer.camera.center = self.prev_camera.center
        self.viewer.camera.zoom = self.prev_camera.zoom

    def set_detected_features(self, det_features: DetectedFeatures):
        """Set the detected features and update the UI"""
        self.det = det_features
        self._intial_det = deepcopy(det_features)
        self._USER_CORRECTED = False

        self.update_features_ui()

    def update_features_ui(self):
        """Update the UI with the detected features"""
        # hide all other layers?
        for layer in self.viewer.layers:
            layer.visible = False

        self._image_layer = self.viewer.add_image(
            self.det.image, name="image", opacity=0.7, blending="additive",
        )

        # add mask to viewer
        self._mask_layer = self.viewer.add_labels(self.det.mask, 
                                                    name="mask", 
                                                    opacity=0.7,
                                                    blending="additive", 
                                                    color=CLASS_COLORS)

        # add points to viewer
        data = []
        for feature in self.det.features:
            x, y = feature.px
            data.append([y, x])

        text = {
            "string": [feature.name for feature in self.det.features],
            "color": "white",
            "translation": np.array([-30, 0]),
        }

        self._features_layer = self.viewer.add_points(
            data,
            name="features",
            text=text,
            size=20,
            edge_width=7,
            edge_width_is_relative=False,
            edge_color="transparent",
            face_color=[feature.color for feature in self.det.features],
            blending="translucent",
        )

        # draw cross hairs
        self._cross_hair_layer = None
        self._draw_crosshairs()

        # set points layer to select mode and active
        self.viewer.layers.selection.active = self._features_layer
        self._features_layer.mode = "select"
        
        # when the point is moved update the feature
        self._features_layer.events.data.connect(self.update_point)
        self.update_info()
            
        # set camera
        self.prev_camera = deepcopy(self.viewer.camera)
        self.viewer.camera.center = [
            0.0,
            self._image_layer.data.shape[0] / 2,
            self._image_layer.data.shape[1] / 2,
        ]
        self.viewer.camera.zoom = 0.7

        if self.det.checkpoint:
            self.label_model.setText(f"Checkpont: {os.path.basename(self.det.checkpoint)}")
        
        napari.utils.notifications.show_info(f"Features ({', '.join([f.name for f in self.det.features])}) Detected")

    def update_info(self):
        """Update the info label with the feature information"""
        if len(self.det.features) > 2:
            self.label_info.setText("Info not available.")
            return
        
        if len(self.det.features) == 1:
            self.label_info.setText(
            f"""{self.det.features[0].name}: {self.det.features[0].px}
            User Corrected: {self._USER_CORRECTED}
            """)
            return
        if len(self.det.features) == 2:
            self.label_info.setText(
                f"""Moving 
                {self.det.features[0].name}: {self.det.features[0].px}
                to 
                {self.det.features[1].name}: {self.det.features[1].px}
                dx={self.det.distance.x*1e6:.2f}um, dy={self.det.distance.y*1e6:.2f}um
                User Corrected: {self._USER_CORRECTED}
                """
                )
            return

    def update_point(self, event):
        """Update the feature when the point is moved"""
        logging.debug(f"{event.source.name} changed its data!")

        layer = self.viewer.layers[f"{event.source.name}"]  # type: ignore

        # get the data
        data = layer.data

        # get which point was moved
        index: list[int] = list(layer.selected_data)  
                
        if len(data) != len(self.det.features):
            # loop backwards to remove the features
            for idx in index[::-1]:
                logging.debug({"msg": "detection_point_deleted",
                               "idx": idx, "data": data[idx], 
                               "feature": self.det.features[idx].name})
                self.det.features.pop(idx)

        else: 
            for idx in index:
                logging.debug({"msg": "detection_point_moved", 
                               "idx": idx, "data": data[idx], 
                               "feature": self.det.features[idx].name})
                
                # update the feature
                self.det.features[idx].px = Point(
                    x=data[idx][1], y=data[idx][0]
                )

        self._draw_crosshairs()
        self._USER_CORRECTED = True
        self.update_info()


    def _draw_crosshairs(self):
        """Draw crosshairs on the image"""

        data = self._features_layer.data

        # for each data point draw two lines from the edge of the image to the point
        line_data, line_colors = [], []
        for idx, point in enumerate(data):
            y, x = point # already flipped
            vline = [[y, 0], [y, self.det.image.data.shape[1]]]
            hline = [[0, x], [self.det.image.data.shape[0], x]]
            
            line_data += [hline, vline]
            color = self.det.features[idx].color
            line_colors += [color, color]
        try:
            self._cross_hair_layer.data = line_data
            self._cross_hair_layer.edge_color = line_colors
        except:
            self._cross_hair_layer = self.viewer.add_shapes(
                data=line_data,
                shape_type="line",
                edge_width=3,
                edge_color=line_colors,
                name="feature_cross_hair",
                opacity=0.7,
                blending="additive",
            )    
    
    def _set_model(self, model: fibsem_model.SegmentationModel):
        self.model = model
        # update model info
        self.label_model.setText(f"Model: {self.model.model.name} \nCheckpont: {os.path.basename(self.model.checkpoint)}")

    def _get_detected_features(self):

        from fibsem import conversions

        for feature in self.det.features:
            feature.feature_m = conversions.image_to_microscope_image_coordinates(
                feature.px, self.det.image.data, self.det.pixelsize
            )

        logging.debug({"msg": "get_detected_features", "detected_features": self.det.to_dict()})

        return self.det


from fibsem.detection.detection import Feature, DetectedFeatures
import fibsem

def main():
    # load model
    checkpoint = "autolamella-mega-20240107.pt"
    model = load_model(checkpoint=checkpoint)
    
    # load image
    image = FibsemImage.load(os.path.join(os.path.dirname(detection.__file__), "test_image_2.tif"))

    pixelsize = image.metadata.pixel_size.x if image.metadata is not None else 25e-9

    # detect features
    features = [detection.LamellaRightEdge(), detection.LandingPost()]
    det = detection.detect_features(
        deepcopy(image.data), model, features=features, pixelsize=pixelsize
    )

    viewer = napari.Viewer(ndisplay=2)
    det_widget_ui = FibsemEmbeddedDetectionUI(
        viewer=viewer, 
        model=model,
        )
    
    det_widget_ui.set_detected_features(det)

    viewer.window.add_dock_widget(
        det_widget_ui, area="right", 
        add_vertical_stretch=False, 
        name=f"OpenFIBSEMv{fibsem.__version__} Feature Detection"
    )
    napari.run()

    det = det_widget_ui.det


if __name__ == "__main__":
    main()


import logging
import napari
import napari.utils.notifications
import numpy as np
from PyQt5 import QtWidgets
from fibsem import config as cfg
from fibsem import constants, conversions, utils, acquire
from fibsem.microscope import FibsemMicroscope
from fibsem.structures import MicroscopeSettings, BeamType, FibsemImage, Point, FibsemStagePosition
from fibsem.ui.qtdesigner_files import FibsemMinimapWidget
import os
from fibsem import patterning
from napari.qt.threading import thread_worker
from fibsem.ui import _stylesheets
from fibsem.imaging import _tile
from scipy.ndimage import median_filter,rotate

from copy import deepcopy

PATH = os.path.join(cfg.DATA_PATH, "tile")
os.makedirs(PATH, exist_ok=True)
from fibsem.ui import utils as ui_utils 

from PyQt5.QtCore import pyqtSignal, pyqtSlot


class FibsemMinimapWidget(FibsemMinimapWidget.Ui_MainWindow, QtWidgets.QMainWindow):
    _stage_position_moved = pyqtSignal(FibsemStagePosition)
    _stage_position_added = pyqtSignal(FibsemStagePosition)
    _update_tile_collection = pyqtSignal(dict)
    _minimap_positions = pyqtSignal(list)
    
    def __init__(
        self,
        microscope: FibsemMicroscope,
        settings: MicroscopeSettings,
        viewer: napari.Viewer,
        parent=None,
    ):
        super(FibsemMinimapWidget, self).__init__(parent=parent)
        self.setupUi(self)

        self.microscope = microscope
        self.settings = settings

        self.parent = parent

        self.viewer = viewer
        self.viewer.window._qt_viewer.dockLayerList.setVisible(False)
        self.viewer.window._qt_viewer.dockLayerControls.setVisible(False)

        self.image = None
        self._image_layer = None
        self._reprojection_layer = None
        self._corr_image_layers = []

        self.positions = []
        self._correlation = {}

        self._tile_info = {}

        self.setup_connections()

        self._update_ui()

    def setup_connections(self):

        self.pushButton_run_tile_collection.clicked.connect(self.run_tile_collection)
        self.pushButton_run_tile_collection.setStyleSheet(_stylesheets._GREEN_PUSHBUTTON_STYLE)
        self.actionLoad_Image.triggered.connect(self.load_image)

        self.comboBox_tile_beam_type.addItems([beam_type.name for beam_type in BeamType])

        self.pushButton_move_to_position.clicked.connect(self._move_position_pressed)

        self.comboBox_tile_position.currentIndexChanged.connect(self._update_current_position_info)
        self.pushButton_update_position.clicked.connect(self._update_position_pressed)
        self.pushButton_update_position.setStyleSheet(_stylesheets._BLUE_PUSHBUTTON_STYLE)
        self.pushButton_remove_position.clicked.connect(self._remove_position_pressed)
        self.pushButton_remove_position.setStyleSheet(_stylesheets._RED_PUSHBUTTON_STYLE)

        self.actionSave_Positions.triggered.connect(self._save_positions_pressed)
        self.actionLoad_Positions.triggered.connect(self._load_positions)

        # checkbox
        self.checkBox_options_move_with_translation.stateChanged.connect(self._update_ui)
        self.checkBox_options_move_with_translation.setVisible(cfg._MINIMAP_MOVE_WITH_TRANSLATION)
        self.label_options_move_with_translate_info.setVisible(cfg._MINIMAP_MOVE_WITH_TRANSLATION)
        self.checkBox_options_acquire_after_movement.setVisible(cfg._MINIMAP_ACQUIRE_AFTER_MOVEMENT)
        self.label_options_header.setVisible(cfg._MINIMAP_ACQUIRE_AFTER_MOVEMENT and cfg._MINIMAP_MOVE_WITH_TRANSLATION)

        # signals
        # self._stage_position_added.connect(self._position_added_callback)
        self._update_tile_collection.connect(self._update_tile_collection_callback)
        self.parent._minimap_signal.connect(self.update_positions_from_parent)


        # pattern overlay
        milling_protocol = self.settings.protocol.get("milling", {})
        milling_patterns = [k for k in milling_protocol if "stages" in milling_protocol[k] or "type" in milling_protocol[k]]
        self.comboBox_pattern_overlay.addItems(milling_patterns)
        if "trench" in milling_patterns:
            self.comboBox_pattern_overlay.setCurrentText("trench")
        elif "lamella" in milling_patterns:
            self.comboBox_pattern_overlay.setCurrentText("lamella")
        self.comboBox_pattern_overlay.currentIndexChanged.connect(self._update_pattern_overlay)
        self.checkBox_pattern_overlay.stateChanged.connect(self._update_pattern_overlay)

        # correlation
        self.actionLoad_Correlation_Image.triggered.connect(self._load_correlation_image)
        self.pushButton_update_correlation_image.clicked.connect(lambda: self._update_correlation_image(None))
        self.comboBox_correlation_selected_layer.currentIndexChanged.connect(self._update_correlation_ui)

        # auto update correlation image
        self.doubleSpinBox_correlation_translation_x.valueChanged.connect(self._update_correlation_data)
        self.doubleSpinBox_correlation_translation_y.valueChanged.connect(self._update_correlation_data)
        self.doubleSpinBox_correlation_scale_x.valueChanged.connect(self._update_correlation_data) 
        self.doubleSpinBox_correlation_scale_y.valueChanged.connect(self._update_correlation_data)
        self.doubleSpinBox_correlation_rotation.valueChanged.connect(self._update_correlation_data)

        self.doubleSpinBox_correlation_translation_x.setKeyboardTracking(False)
        self.doubleSpinBox_correlation_translation_y.setKeyboardTracking(False)
        self.doubleSpinBox_correlation_scale_x.setKeyboardTracking(False)
        self.doubleSpinBox_correlation_scale_y.setKeyboardTracking(False)
        self.doubleSpinBox_correlation_rotation.setKeyboardTracking(False)

        self.lineEdit_tile_path.setText(self.settings.image.path)

        # gridbar

        self.checkBox_gridbar.stateChanged.connect(self._gridbar_set)
        self.label_gb_spacing.setVisible(False)
        self.label_gb_width.setVisible(False)
        self.doubleSpinBox_gb_spacing.setVisible(False)
        self.doubleSpinBox_gb_width.setVisible(False)

        self.doubleSpinBox_gb_spacing.valueChanged.connect(self._update_gridbar)
        self.doubleSpinBox_gb_width.valueChanged.connect(self._update_gridbar)
        self.doubleSpinBox_gb_spacing.setKeyboardTracking(False)
        self.doubleSpinBox_gb_width.setKeyboardTracking(False)


    def update_positions_from_parent(self, positions):
        
        if positions is not None:
            self.positions = positions

        self._update_position_info()
        self._update_viewer()


    def run_tile_collection(self):

        logging.info("running tile collection")

        beam_type = BeamType[self.comboBox_tile_beam_type.currentText()]
        grid_size = self.doubleSpinBox_tile_grid_size.value() * constants.MICRO_TO_SI
        tile_size = self.doubleSpinBox_tile_tile_size.value() * constants.MICRO_TO_SI
        resolution = int(self.spinBox_tile_resolution.value())
        dwell_time = self.doubleSpinBox_tile_dwell_time.value() * constants.MICRO_TO_SI
        cryo = self.checkBox_tile_autogamma.isChecked()
        autocontrast = self.checkBox_tile_autogamma.isChecked()
        path = self.lineEdit_tile_path.text()
        filename = self.lineEdit_tile_filename.text()

        self._tile_info["grid_size"] = grid_size
        self._tile_info["tile_size"] = tile_size
        self._tile_info["resolution"] = resolution
        self._tile_info["beam_type"] = beam_type
                
        self.settings.image.dwell_time = dwell_time 
        self.settings.image.hfw = tile_size
        self.settings.image.beam_type = beam_type
        self.settings.image.resolution = [resolution, resolution]
        self.settings.image.autocontrast = autocontrast
        self.settings.image.save = True
        self.settings.image.path = path
        self.settings.image.filename = filename

        if self.settings.image.filename == "":
            napari.utils.notifications.show_error(f"Please enter a filename for the image")
            return
        
        # ui feedback
        self.pushButton_run_tile_collection.setStyleSheet(_stylesheets._ORANGE_PUSHBUTTON_STYLE)
        self.pushButton_run_tile_collection.setText("Running Tile Collection...")

        worker = self._run_tile_collection_thread(
            microscope=self.microscope, settings=self.settings, 
            grid_size=grid_size, tile_size=tile_size, 
            overlap=0, cryo=cryo)

        worker.finished.connect(self._tile_collection_finished)
        worker.start()


    def _gridbar_set(self):

        if self.checkBox_gridbar.isChecked():

            grid_shape = self.image.data.shape
            arr = np.zeros(shape=grid_shape, dtype=np.uint8)

            pixelsize = self.image.metadata.pixel_size.x

            # create grid, grid bars thickness = 10px
            BAR_THICKNESS_PX = int(5 * constants.MICRO_TO_SI / pixelsize)
            BAR_SPACING_PX = int(50 * constants.MICRO_TO_SI / pixelsize)   
            for i in range(0, arr.shape[0], BAR_SPACING_PX ):
                arr[i:i+BAR_THICKNESS_PX, :] = 255
                arr[:, i:i+BAR_THICKNESS_PX] = 255

            gridbar_image = FibsemImage(data=arr)

            self._update_correlation_image(gridbar_image, gridbar=True)

            self.label_gb_spacing.setVisible(True)
            self.label_gb_width.setVisible(True)
            self.doubleSpinBox_gb_spacing.setVisible(True)
            self.doubleSpinBox_gb_width.setVisible(True)

            self.doubleSpinBox_gb_spacing.setValue(50)
            self.doubleSpinBox_gb_width.setValue(5)
            

        else:
            
            layer_to_remove = ""
            for layer in self.viewer.layers:
                if "gridbar" in layer.name:
                    layer_to_remove = layer.name
            
            self.label_gb_spacing.setVisible(False)
            self.label_gb_width.setVisible(False)
            self.doubleSpinBox_gb_spacing.setVisible(False)
            self.doubleSpinBox_gb_width.setVisible(False)

            self.viewer.layers.remove(layer_to_remove)
            self.comboBox_correlation_selected_layer.currentIndexChanged.disconnect()
            self.comboBox_correlation_selected_layer.clear()
            self.comboBox_correlation_selected_layer.addItems([layer.name for layer in self.viewer.layers if "correlation-image" in layer.name ])
            self.comboBox_correlation_selected_layer.currentIndexChanged.connect(self._update_correlation_ui)


    def _update_gridbar(self):

        pixel_size = self.image.metadata.pixel_size.x

        print(f'pixel size: {pixel_size}')


        BAR_THICKNESS_PX = int(self.doubleSpinBox_gb_width.value() * constants.MICRO_TO_SI / pixel_size)
        BAR_SPACING_PX = int(self.doubleSpinBox_gb_spacing.value() * constants.MICRO_TO_SI / pixel_size)

        gridbar_layer = ''

        for layer in self.viewer.layers:
            if "gridbar" in layer.name:
                gridbar_layer = layer.name
        
        grid_shape = self.viewer.layers[gridbar_layer].data.shape

        arr = np.zeros(shape=grid_shape, dtype=np.uint8)
        for i in range(0, arr.shape[0], BAR_SPACING_PX ):
            arr[i:i+BAR_THICKNESS_PX, :] = 255
            arr[:, i:i+BAR_THICKNESS_PX] = 255

        self.viewer.layers[gridbar_layer].data = arr




    def _tile_collection_finished(self):

        napari.utils.notifications.show_info(f"Tile collection finished.")
        self._update_viewer(self.image)
        self.pushButton_run_tile_collection.setStyleSheet(_stylesheets._GREEN_PUSHBUTTON_STYLE)
        self.pushButton_run_tile_collection.setText("Run Tile Collection")

    
    @thread_worker
    def _run_tile_collection_thread(self, microscope: FibsemMicroscope, settings:MicroscopeSettings, 
        grid_size: float, tile_size:float, overlap:float=0, cryo: bool=True):

        self.image = _tile._tile_image_collection_stitch(
            microscope=microscope, settings=settings, 
            grid_size=grid_size, tile_size=tile_size, 
            overlap=overlap, cryo=cryo, parent_ui=self)

    def _update_tile_collection_callback(self, ddict):
        
        # msg = f"{ddict['msg']} ({ddict['i']+1}/{ddict['n_rows']}, {ddict['j']+1}/{ddict['n_cols']})"
        msg = f"{ddict['msg']} ({ddict['counter']}/{ddict['total']})"
        logging.info(msg)
        napari.utils.notifications.show_info(msg)

        if ddict["image"] is not None:

            arr = median_filter(ddict["image"], size=3)
            try:
                self._image_layer.data = arr
            except:
                self._image_layer = self.viewer.add_image(arr, name="overview-image", colormap="gray", blending="additive")

    def load_image(self):

        path = ui_utils._get_file_ui( msg="Select image to load", path=cfg.DATA_TILE_PATH, _filter="Image Files (*.tif *.tiff)", parent=self)

        if path == "":
            napari.utils.notifications.show_info(f"No file selected..")
            return

        image = FibsemImage.load(path)

        self._tile_info["tile_size"] = image.metadata.image_settings.hfw
        self._tile_info["resolution"] = image.metadata.image_settings.resolution[0]
        self._tile_info["beam_type"] = image.metadata.image_settings.beam_type

        self._update_viewer(image)

    def _update_ui(self):

        _image_loaded = self.image is not None
        self.pushButton_update_position.setEnabled(_image_loaded)
        self.pushButton_remove_position.setEnabled(_image_loaded)
        self.pushButton_move_to_position.setEnabled(_image_loaded)

        if _image_loaded:
            self.label_instructions.setText("Alt+Click to Add a position, Shift+Click to Update a position \nor Double Click to Move the Stage...")
        else:
            self.label_instructions.setText("Please take or load an overview image...")

        _positions_loaded = len(self.positions) > 0
        self.pushButton_move_to_position.setEnabled(_positions_loaded)

        _MOVE_WITH_TRANSLATION = self.checkBox_options_move_with_translation.isChecked()
        self.label_translation_x.setVisible(_MOVE_WITH_TRANSLATION)
        self.label_translation_y.setVisible(_MOVE_WITH_TRANSLATION)
        self.label_translation_z.setVisible(_MOVE_WITH_TRANSLATION)
        self.doubleSpinBox_translation_x.setVisible(_MOVE_WITH_TRANSLATION)
        self.doubleSpinBox_translation_y.setVisible(_MOVE_WITH_TRANSLATION)
        self.doubleSpinBox_translation_z.setVisible(_MOVE_WITH_TRANSLATION)

    def _update_viewer(self, image: FibsemImage =  None):

        if image is not None:
            self.image = image
        
            arr = median_filter(self.image.data, size=3)
            try:
                self._image_layer.data = arr
            except:
                self._image_layer = self.viewer.add_image(arr, name="overview-image", colormap="gray", blending="additive")

            # draw a point on the image at center
            ui_utils._draw_crosshair(viewer=self.viewer,eb_image= self.image, ib_image= self.image,is_checked=True) 

            self._image_layer.mouse_drag_callbacks.clear()
            self._image_layer.mouse_double_click_callbacks.clear()
            self._image_layer.mouse_drag_callbacks.append(self._on_click)
            self._image_layer.mouse_double_click_callbacks.append(self._on_double_click)
    

        self._draw_positions() # draw the reprojected positions on the image

        self.viewer.layers.selection.active = self._image_layer

        self._update_ui()


    def _update_region(self, position:FibsemStagePosition):
    

        resolution = self._tile_info["resolution"]
        self.settings.image.resolution = [resolution, resolution]
        self.settings.image.beam_type = self._tile_info["beam_type"]
        self.settings.image.hfw = self._tile_info["tile_size"]
        self.settings.image.save = False

        # TODO: this assumes the image is taken with the same settings as the tile collection
        self.image = _tile._update_image_region(self.microscope, self.settings.image, self.image, position)

        self._update_viewer(self.image)

    def get_data_from_coord(self, coords: tuple) -> tuple:
        # check inside image dimensions, (y, x)
        shape = self.image.data.shape[0], self.image.data.shape[1]

        if (coords[0] > 0 and coords[0] < shape[0]) and (
            coords[1] > 0 and coords[1] < shape[1]
        ):
            return True
        else:
            return False


    def _validate_mouse_click(self, layer, event):
        """validate if click is inside image, and convert to microscope coords"""                
        # get coords
        coords = layer.world_to_data(event.position)

        _inside_image = self.get_data_from_coord(coords)

        if _inside_image is False:
            napari.utils.notifications.show_warning(
                f"Clicked outside image dimensions. Please click inside the image to move."
            )
            return False, False

        point = conversions.image_to_microscope_image_coordinates(
            Point(x=coords[1], y=coords[0]), self.image.data, self.image.metadata.pixel_size.x,
        )

        return coords, point

    def _on_click(self, layer, event):
        
        # left click + (alt or shift)
        if event.button != 1 or ('Alt' not in event.modifiers and 'Shift' not in event.modifiers):
            return 

        coords, point = self._validate_mouse_click(layer, event)

        if point is False: # clicked outside image
            return

        _new_position = self.microscope.project_stable_move( 
                    dx=point.x, dy=point.y, 
                    beam_type=self.image.metadata.image_settings.beam_type, 
                    base_position=self.image.metadata.microscope_state.stage_position)            
       
        if 'Shift' in event.modifiers:
            idx = self.comboBox_tile_position.currentIndex()
            _name = self.positions[idx].name
            _new_position.name = _name
            self.positions[idx] = _new_position
        elif 'Alt' in event.modifiers:
            _new_position.name = f"Position {len(self.positions)+1:02d}" 
            self.positions.append(_new_position)
            self._stage_position_added.emit(_new_position)


        # we could save this position as well, use it to pre-select a bunch of lamella positions?
        self._update_position_info()
        self._update_viewer()


        self._minimap_positions.emit(self.positions)

    def _on_double_click(self, layer, event):
        
        if event.button !=1:
            return
        coords, point = self._validate_mouse_click(layer, event)

        if point is False: # clicked outside image
            return

        _new_position = self.microscope.project_stable_move( 
            dx=point.x, dy=point.y, 
            beam_type=self.image.metadata.image_settings.beam_type, 
            base_position=self.image.metadata.microscope_state.stage_position)   


        _MOVE_WITH_TRANSLATION = self.checkBox_options_move_with_translation.isChecked()
        if _MOVE_WITH_TRANSLATION:
            dx = self.doubleSpinBox_translation_x.value() * constants.MILLI_TO_SI
            dy = self.doubleSpinBox_translation_y.value() * constants.MILLI_TO_SI
            dz = self.doubleSpinBox_translation_z.value() * constants.MILLI_TO_SI
            translation = FibsemStagePosition(x=dx, y=dy, z=dz)
            logging.info(f"[NOT ENABLED] Moving to position with translation: {translation}")
            napari.utils.notifications.show_warning(f" [NOT ENABLED] Moving to position with translation: {translation}")

            if cfg._MINIMAP_MOVE_WITH_TRANSLATION:
                _new_position += translation

        self._move_to_position(_new_position)
        self._minimap_positions.emit(self.positions)

    def _update_current_position_info(self):

        idx = self.comboBox_tile_position.currentIndex()
        if idx != -1 and len(self.positions) > 0:
            self.lineEdit_tile_position_name.setText(self.positions[idx].name)
            self.pushButton_move_to_position.setText(f"Move to {self.positions[idx].name}")
        else:
            self.lineEdit_tile_position_name.setText("")
        

    def _update_position_info(self):
        idx = self.comboBox_tile_position.currentIndex()
        self.comboBox_tile_position.clear()
        self.comboBox_tile_position.addItems([pos.name for pos in self.positions])
        if idx != -1:
            self.comboBox_tile_position.setCurrentIndex(idx)

        _positions_added = len(self.positions) > 0
        self.pushButton_move_to_position.setEnabled(_positions_added)
        self.pushButton_remove_position.setEnabled(_positions_added)
        self.pushButton_update_position.setEnabled(_positions_added)

        msg = ""
        for pos in self.positions:
            msg += f"{pos.name}:\t x:{pos.x*1e3:.2f}mm, y:{pos.y*1e3:.2f}mm, z:{pos.z*1e3:.2f}m\n"
        self.label_position_info.setText(msg)


    def _update_position_pressed(self):
        logging.info(f"Updating position...")
        _position = self.positions[self.comboBox_tile_position.currentIndex()]
        
        name = self.lineEdit_tile_position_name.text()
        if name == "":
            napari.utils.notifications.show_info(f"Please enter a name for the position")
            return

        _position.name = name 
        self._update_position_info()
        self._update_viewer()

    def _remove_position_pressed(self):
        logging.info(f"Removing position...")
        _position = self.positions[self.comboBox_tile_position.currentIndex()]
        self.positions.remove(_position)
        self._minimap_positions.emit(self.positions)
        self._update_position_info()
        self._update_viewer()
        


    def _move_position_pressed(self):

        logging.info(f"Moving to position...")
        _position = self.positions[self.comboBox_tile_position.currentIndex()]
        logging.info(f"Moving To: {_position}")
        self._move_to_position(_position)

    def _move_to_position(self, _position:FibsemStagePosition)->None:
        self.microscope.safe_absolute_stage_movement(_position)
        self._stage_position_moved.emit(_position)
        if self.checkBox_options_acquire_after_movement.isChecked():
            self._update_region(_position)
        self._update_viewer()


    def _load_positions(self):

        logging.info(f"Loading Positions...")
        path = ui_utils._get_file_ui( msg="Select a position file to load", 
            path=self.settings.image.path, 
            _filter= "*yaml", 
            parent=self)

        if path == "":
            napari.utils.notifications.show_info(f"No file selected..")
            return

        pdict = utils.load_yaml(path)
        
        positions = [FibsemStagePosition.from_dict(p) for p in pdict]
        # self.positions = self.positions + positions # append? or overwrite
        # overwrite 
        self.positions = positions
        self._minimap_positions.emit(self.positions)

        self._update_position_info()
        self._update_viewer()

    def _save_positions_pressed(self):
        
        logging.info(f"Saving Positions...")
        
        path = ui_utils._get_save_file_ui(msg = "Select a file to save the positions to",
            path = self.settings.image.path,
            _filter= "*yaml",
            parent=self,
        )

        if path == "":
            napari.utils.notifications.show_info(f"No file selected, not saving.")
            return

        # save the positions
        pdict = [p.to_dict() for p in self.positions]
        utils.save_yaml(path, pdict)

        napari.utils.notifications.show_info(f"Saved positions to {path}")

    def _update_pattern_overlay(self):

        self._draw_positions()


    def _draw_positions(self):
        
        logging.info(f"Drawing Reprojected Positions...")
        current_position = deepcopy(self.microscope.get_stage_position())
        current_position.name = f"Current Position"

        if self.image:
            
            drawn_positions = self.positions + [current_position]
            points = _tile._reproject_positions(self.image, drawn_positions)

            data = []
            for pt in points:
                # reverse to list
                data.append([pt.y, pt.x])

            colors = ["lime"] * (len(drawn_positions)-1)
            colors.append("yellow")

            colors_rgba = [[0,1,0,1] for _ in range(len(drawn_positions)-1)]
            colors_rgba.append([1,1,0,1]) # yellow

            text = {
                "string": [pos.name for pos in drawn_positions],
                "color": colors, # TODO: separate colour for current position
                "translation": np.array([-50, 0]),
            }

            if self._reprojection_layer is None:
            
                 self._reprojection_layer = self.viewer.add_points(
                    data,
                    name="Positions",
                    text=text,
                    size=60,
                    edge_width=7,
                    edge_width_is_relative=False,
                    edge_color="transparent",
                    face_color="lime",
                    blending="translucent",
                    symbol="cross",
                )
            else:
                self._reprojection_layer.data = data
                self._reprojection_layer.text = text

            self._reprojection_layer.face_color= colors_rgba


            _SHOW_PATTERNS: bool = self.checkBox_pattern_overlay.isChecked()
            if _SHOW_PATTERNS: # TODO: this is very slow, need to speed up, too many pattern redraws
                points = [conversions.image_to_microscope_image_coordinates(Point(x=coords[1], y=coords[0]), self.image.data, self.image.metadata.pixel_size.x ) for coords in data[:-1]]
                pattern = self.comboBox_pattern_overlay.currentText() 
                
                milling_stages = [patterning.get_milling_stages(pattern, self.settings.protocol["milling"], Point(point.x, point.y))[0] for point in points]
                
                for stage, pos in zip(milling_stages, drawn_positions[:-1]):
                    stage.name = pos.name

                ui_utils._draw_patterns_in_napari(self.viewer, 
                    ib_image=self.image, 
                    eb_image=None, 
                    milling_stages = milling_stages)
            else:
                ui_utils._remove_all_layers(viewer=self.viewer)

        self.viewer.layers.selection.active = self._image_layer


    def _load_correlation_image(self):

        # load another image
        path = ui_utils._get_file_ui( msg="Select image to load", 
        path=cfg.DATA_TILE_PATH, _filter="Image Files (*.tif *.tiff)", parent=self)

        if path == "":
            napari.utils.notifications.show_info(f"No file selected..")
            return

        image = FibsemImage.load(path)    

        self._update_correlation_image(image)


    def _update_correlation_image(self, image: FibsemImage = None,gridbar:bool=False):

        if image is not None:
            
            _basename = f"correlation-image" if not gridbar else f"gridbar-image"
            idx = 1
            _name = f"{_basename}-{idx:02d}"
            while _name in self.viewer.layers:
                idx+=1
                _name = f"{_basename}-{idx:02d}"

            # if grid bar in _name, idx = 3
            if gridbar:
                idx = 3

            COLOURS = ["green", "cyan", "magenta", "red", "yellow"]

            self._corr_image_layers.append(
                self.viewer.add_image(image.data, 
                name=_name, colormap=COLOURS[idx%len(COLOURS)], 
                blending="translucent", opacity=0.2)
            )

            self._correlation[_name] = deepcopy([0, 0, 1.0, 1.0, 0])


            self.comboBox_correlation_selected_layer.currentIndexChanged.disconnect()
            idx = self.comboBox_correlation_selected_layer.currentIndex()
            self.comboBox_correlation_selected_layer.clear()
            self.comboBox_correlation_selected_layer.addItems([layer.name for layer in self.viewer.layers if "correlation-image" in layer.name or "gridbar-image" in layer.name])
            if idx != -1:
                self.comboBox_correlation_selected_layer.setCurrentIndex(idx)
            self.comboBox_correlation_selected_layer.currentIndexChanged.connect(self._update_correlation_ui)
            
            self.viewer.layers.selection.active = self._image_layer
    
    # do this when image selected is changed
    def _update_correlation_ui(self):

        # set ui
        layer_name = self.comboBox_correlation_selected_layer.currentText()
        if layer_name == "":
            napari.utils.notifications.show_info(f"Please select a layer to correlate with  update data...")
            return

        tx, ty, sx, sy, r = self._correlation[layer_name]

        self.doubleSpinBox_correlation_translation_x.setValue(tx)
        self.doubleSpinBox_correlation_translation_y.setValue(ty)
        self.doubleSpinBox_correlation_scale_x.setValue(sx)
        self.doubleSpinBox_correlation_scale_y.setValue(sy)
        self.doubleSpinBox_correlation_rotation.setValue(r)

        self._update_correlation_data()

    # do this when parameters change
    def _update_correlation_data(self):

        # select layer
        layer_name = self.comboBox_correlation_selected_layer.currentText()
        if layer_name == "":
            napari.utils.notifications.show_info(f"Please select a layer to correlate with update ui...")
            return

        tx, ty = self.doubleSpinBox_correlation_translation_x.value(), self.doubleSpinBox_correlation_translation_y.value()
        sx, sy = self.doubleSpinBox_correlation_scale_x.value(), self.doubleSpinBox_correlation_scale_y.value()
        r = self.doubleSpinBox_correlation_rotation.value()

        if sx == 0 or sy == 0:
            sx =1
            sy =1

        angle = np.deg2rad(r)

        rows = self.viewer.layers[layer_name].data.shape[0]*0.5 
        cols = self.viewer.layers[layer_name].data.shape[1]*0.5


        # the proof for this is marvelous but i dont have enough space in the comments

        new_x = int(np.cos(angle) * rows - np.sin(angle) * cols) 
        new_y =  int(np.sin(angle) * rows + np.cos(angle) * cols) 

        trans_x = new_x -rows 
        trans_y = new_y -cols 

        # translation for central rotation + translation
        corrected_tx = tx - trans_y
        corrected_ty = -ty - trans_x


        self._correlation[layer_name] = deepcopy([tx, ty, sx, sy, r])

        # apply to selected layer
        self.viewer.layers[layer_name].translate = [corrected_ty, corrected_tx]
        self.viewer.layers[layer_name].scale = [sy, sx]
        self.viewer.layers[layer_name].rotate = r

# TODO: update layer name, set from file?
# TODO: set combobox to all images in viewer 



def main():

    viewer = napari.Viewer(ndisplay=2)
    microscope, settings = utils.setup_session()
    minimap_widget = FibsemMinimapWidget(microscope, settings, viewer=viewer)
    viewer.window.add_dock_widget(
        minimap_widget, area="right", add_vertical_stretch=False, name="OpenFIBSEM Minimap"
    )
    napari.run()


if __name__ == "__main__":
    main()


# -*- coding: utf-8 -*-

# Form implementation generated from reading ui file 'FibsemMicroscopeConfigurationWidget.ui'
#
# Created by: PyQt5 UI code generator 5.15.9
#
# WARNING: Any manual changes made to this file will be lost when pyuic5 is
# run again.  Do not edit this file unless you know what you are doing.


from PyQt5 import QtCore, QtGui, QtWidgets


class Ui_MainWindow(object):
    def setupUi(self, MainWindow):
        MainWindow.setObjectName("MainWindow")
        MainWindow.resize(632, 930)
        self.centralwidget = QtWidgets.QWidget(MainWindow)
        self.centralwidget.setObjectName("centralwidget")
        self.gridLayout = QtWidgets.QGridLayout(self.centralwidget)
        self.gridLayout.setObjectName("gridLayout")
        self.label_configuration_title = QtWidgets.QLabel(self.centralwidget)
        font = QtGui.QFont()
        font.setPointSize(12)
        font.setBold(True)
        font.setWeight(75)
        self.label_configuration_title.setFont(font)
        self.label_configuration_title.setObjectName("label_configuration_title")
        self.gridLayout.addWidget(self.label_configuration_title, 0, 0, 1, 1)
        MainWindow.setCentralWidget(self.centralwidget)
        self.menubar = QtWidgets.QMenuBar(MainWindow)
        self.menubar.setGeometry(QtCore.QRect(0, 0, 632, 22))
        self.menubar.setObjectName("menubar")
        self.menuFile = QtWidgets.QMenu(self.menubar)
        self.menuFile.setObjectName("menuFile")
        MainWindow.setMenuBar(self.menubar)
        self.statusbar = QtWidgets.QStatusBar(MainWindow)
        self.statusbar.setObjectName("statusbar")
        MainWindow.setStatusBar(self.statusbar)
        self.actionLoad_Configuration = QtWidgets.QAction(MainWindow)
        self.actionLoad_Configuration.setObjectName("actionLoad_Configuration")
        self.actionSave_Configuration = QtWidgets.QAction(MainWindow)
        self.actionSave_Configuration.setObjectName("actionSave_Configuration")
        self.menuFile.addAction(self.actionLoad_Configuration)
        self.menuFile.addAction(self.actionSave_Configuration)
        self.menubar.addAction(self.menuFile.menuAction())

        self.retranslateUi(MainWindow)
        QtCore.QMetaObject.connectSlotsByName(MainWindow)

    def retranslateUi(self, MainWindow):
        _translate = QtCore.QCoreApplication.translate
        MainWindow.setWindowTitle(_translate("MainWindow", "MainWindow"))
        self.label_configuration_title.setText(_translate("MainWindow", "OpenFIBSEM: Microscope Configuration"))
        self.menuFile.setTitle(_translate("MainWindow", "File"))
        self.actionLoad_Configuration.setText(_translate("MainWindow", "Load Configuration"))
        self.actionSave_Configuration.setText(_translate("MainWindow", "Save Configuration"))
